# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_data_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
from decoder_for_shared_package import *
import typeset



def decode_data_content(xml, nsmap):
    """Decodes an instance of the following type: data content.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataContent

    """
    decodings = [
        ('aggregation', False, 'unicode', 'child::cim:aggregation'),
        ('frequency', False, 'unicode', 'child::cim:frequency/@value'),
        ('topic', False, decode_data_topic, 'child::cim:topic'),
    ]

    return set_attributes(typeset.data.DataContent(), xml, nsmap, decodings)


def decode_data_distribution(xml, nsmap):
    """Decodes an instance of the following type: data distribution.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataDistribution

    """
    decodings = [
        ('access', False, 'unicode', '@distributionAccess'),
        ('format', False, 'unicode', 'child::cim:distributionFormat/@value'),
        ('responsible_party', False, decode_responsible_party, 'child::cim:responsibleParty'),
    ]

    return set_attributes(typeset.data.DataDistribution(), xml, nsmap, decodings)


def decode_data_extent(xml, nsmap):
    """Decodes an instance of the following type: data extent.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataExtent

    """
    decodings = [
        ('geographical', False, decode_data_extent_geographical, 'child::gmd:geographicElement'),
        ('temporal', False, decode_data_extent_temporal, 'child::gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod'),
    ]

    return set_attributes(typeset.data.DataExtent(), xml, nsmap, decodings)


def decode_data_extent_geographical(xml, nsmap):
    """Decodes an instance of the following type: data extent geographical.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataExtentGeographical

    """
    decodings = [
        ('east', False, 'float', 'child::gmd:EX_GeographicBoundingBox/gmd:eastBoundLongitude/gco:Decimal'),
        ('north', False, 'float', 'child::gmd:EX_GeographicBoundingBox/gmd:northBoundLatitude/gco:Decimal'),
        ('south', False, 'float', 'child::gmd:EX_GeographicBoundingBox/gmd:southBoundLatitude/gco:Decimal'),
        ('west', False, 'float', 'child::gmd:EX_GeographicBoundingBox/gmd:westBoundLongitude/gco:Decimal'),
    ]

    return set_attributes(typeset.data.DataExtentGeographical(), xml, nsmap, decodings)


def decode_data_extent_temporal(xml, nsmap):
    """Decodes an instance of the following type: data extent temporal.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataExtentTemporal

    """
    decodings = [
        ('begin', False, 'datetime.date', 'child::gml:beginPosition'),
        ('end', False, 'datetime.date', 'child::gml:endPosition'),
        ('time_interval', False, decode_data_extent_time_interval, 'child::gml:timeInterval'),
    ]

    return set_attributes(typeset.data.DataExtentTemporal(), xml, nsmap, decodings)


def decode_data_extent_time_interval(xml, nsmap):
    """Decodes an instance of the following type: data extent time interval.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataExtentTimeInterval

    """
    decodings = [
        ('factor', False, 'int', '@factor'),
        ('radix', False, 'int', '@radix'),
        ('unit', False, 'unicode', '@unit'),
    ]

    return set_attributes(typeset.data.DataExtentTimeInterval(), xml, nsmap, decodings)


def decode_data_hierarchy_level(xml, nsmap):
    """Decodes an instance of the following type: data hierarchy level.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataHierarchyLevel

    """
    decodings = [
        ('is_open', False, 'bool', 'child::cim:hierarchyLevelName/@open'),
        ('name', False, 'unicode', 'child::cim:hierarchyLevelName/@value'),
        ('value', False, 'unicode', 'child::cim:hierarchyLevelValue'),
    ]

    return set_attributes(typeset.data.DataHierarchyLevel(), xml, nsmap, decodings)


def decode_data_object(xml, nsmap):
    """Decodes an instance of the following type: data object.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataObject

    """
    decodings = [
        ('acronym', False, 'unicode', 'child::cim:acronym'),
        ('citations', True, decode_citation, '//cim:citation[not(cim:citation)]'),
        ('content', True, decode_data_content, 'child::cim:content'),
        ('data_status', False, 'unicode', 'self::cim:dataObject/@dataStatus'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('distribution', False, decode_data_distribution, 'child::cim:distribution'),
        ('extent', False, decode_data_extent, 'child::cim:extent'),
        ('hierarchy_level', False, decode_data_hierarchy_level, 'self::cim:dataObject'),
        ('keyword', False, 'unicode', 'child::cim:keyword'),
        ('meta', False, decode_doc_meta_info, 'self::cim:dataObject'),
        ('properties', True, decode_data_property, 'child::cim:dataProperty/cim:dataProperty'),
        ('purpose', False, 'unicode', 'self::cim:dataObject/@purpose'),
        ('purpose', False, 'unicode', 'self::cim:dataObject/@purpose'),
        ('storage', True, decode_data_storage_ip, 'child::cim:storage/cim:ipStorage'),
    ]

    return set_attributes(typeset.data.DataObject(), xml, nsmap, decodings)


def decode_data_property(xml, nsmap):
    """Decodes an instance of the following type: data property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataProperty

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.data.DataProperty(), xml, nsmap, decodings)


def decode_data_restriction(xml, nsmap):
    """Decodes an instance of the following type: data restriction.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataRestriction

    """
    decodings = [
    ]

    return set_attributes(typeset.data.DataRestriction(), xml, nsmap, decodings)


def decode_data_storage(xml, nsmap):
    """Decodes an instance of the following type: data storage.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataStorage

    """
    decodings = [
    ]

    return set_attributes(typeset.data.DataStorage(), xml, nsmap, decodings)


def decode_data_storage_db(xml, nsmap):
    """Decodes an instance of the following type: data storage db.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataStorageDb

    """
    decodings = [
    ]

    return set_attributes(typeset.data.DataStorageDb(), xml, nsmap, decodings)


def decode_data_storage_file(xml, nsmap):
    """Decodes an instance of the following type: data storage file.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataStorageFile

    """
    decodings = [
    ]

    return set_attributes(typeset.data.DataStorageFile(), xml, nsmap, decodings)


def decode_data_storage_ip(xml, nsmap):
    """Decodes an instance of the following type: data storage ip.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataStorageIp

    """
    decodings = [
        ('file_name', False, 'unicode', 'child::cim:fileName'),
        ('format', False, 'unicode', 'child::cim:dataFormat/@value'),
    ]

    return set_attributes(typeset.data.DataStorageIp(), xml, nsmap, decodings)


def decode_data_topic(xml, nsmap):
    """Decodes an instance of the following type: data topic.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.data.DataTopic

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('unit', False, 'unicode', 'child::cim:unit/@value'),
    ]

    return set_attributes(typeset.data.DataTopic(), xml, nsmap, decodings)





"""
.. module:: loader.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Loads CMIP6 specialization from file system.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import imp
import os



def get_realm_specializations(input_dir, realm_name):
    """Returns specialization modules organized by type.

    :param str input_dir: Directory within which modules reside.
    :param str realm_name: Name of realm being processed.

    """
    def get_module(modules, name):
        for module in modules:
            if module.__name__ == name:
                return module


    def is_target(filename):
        """Returns flag indicating whether a module is a specialization target or not.

        """
        return not filename.startswith('_') and \
               filename.endswith('.py') and \
               filename.startswith(realm_name)


    # Load specialization modules.
    modules = sorted([i for i in os.listdir(input_dir) if is_target(i)])
    modules = [os.path.join(input_dir, m) for m in modules]
    modules = [(m.split("/")[-1].split(".")[0], m) for m in modules]
    modules = [imp.load_source(name, fpath) for name, fpath in modules]
    if not modules:
        raise KeyError("Realm specializations not found")

    # Set specializations.
    realm = get_module(modules, realm_name)
    grid = get_module(modules, "{}_grid".format(realm_name))
    key_properties = get_module(modules, "{}_key_properties".format(realm_name))
    processes = [get_module(modules, p) for p in realm.PROCESSES]

    return realm, grid, key_properties, processes



"""A realm key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean key properties'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in ocean',
    'properties': [
        ('model_name', 'str', '1.1',
            'Name of ocean model code (NEMO 3.6, MOM 5.0,...)'),
        ('model_family', 'ENUM:model_family_types', '1.1',
            'Type of ocean model.'),
        ('basic_approximations', 'ENUM:ocean_basic_approx_types', '1.N',
            'Basic approximations made in the ocean.',),
        ('prognostic_variables', 'ENUM:prognostic_vars_types', '1.N',
            'List of prognostic variables in the ocean component.'),
        ]
    }

DETAILS['toplevel:seawater_properties'] = {
    'description': 'Physical properties of seawater in ocean',
    'properties' : [
        ('eos_type', 'ENUM:seawater_eos_types', '1.1',
            'Type of EOS for sea water'),
        ('eos_functional_temp', 'ENUM:seawater_eos_func_temp', '1.1',
            'Temperature used in EOS for sea water'),
        ('eos_functional_salt', 'ENUM:seawater_eos_func_salt', '1.1',
            'Salinity used in EOS for sea water'),
        ('eos_functional_depth', 'ENUM:seawater_eos_func_depth', '1.1',
            'Depth or pressure used in EOS for sea water ?'),
        ('ocean_freezing_point', 'ENUM:seawater_freezing_point', '1.1',
            'Equation used to compute the freezing point (in deg C) of seawater, as a function of salinity and pressure'),
        ('ocean_specific_heat', 'float', '1.1',
            'Specific heat in ocean (cpocean) in J/(kg K)'),
        ('ocean_reference_density', 'float', '1.1',
            'Boussinesq reference density (rhozero) in kg / m3'),
        ]
    }

DETAILS['toplevel:bathymetry'] = {
    'description': 'Properties of bathymetry in ocean',
    'properties' : [
        ('reference_dates', 'ENUM:bathymetry_ref_dates', '1.1',
            'Reference date of bathymetry'),
        ('type', 'bool', '1.1',
            'Is the bathymetry fixed in time in the ocean ?'),
        ('ocean_smoothing', 'str', '1.1',
            'Describe any smoothing or hand editing of bathymetry in ocean'),
        ('source', 'str', '1.1',
            'Describe source of bathymetry in ocean'),
        ]
    }

DETAILS['toplevel:nonoceanic_waters'] = {
    'description': 'Non oceanic waters treatement in ocean',
    'properties': [
        ('isolated_seas','str', '0.1',
            'Describe if/how isolated seas is performed'),
        ('river_mouth','str', '0.1',
            'Describe if/how river mouth mixing or estuaries specific treatment is performed'),
       ]
    }

DETAILS['software_properties'] = {
    'description': 'Software properties of ocean code',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','str', '0.N',
            "Code language(s)."),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: RESOLUTION: The resolution of the grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': 'Resolution in the ocean grid',
    'properties': [
        ('name', 'str', '1.1',
             "This is a string usually used by the modelling group to describe the resolution of this grid, e.g. ORCA025, N512L180, T512L70 etc."),
        ('canonical_horizontal_resolution', 'str', '0.1',
             "Expression quoted for gross comparisons of resolution, eg. 50km or 0.1 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '0.1',
             "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ('number_of_vertical_levels', 'int', '0.1',
             "Number of vertical levels resolved on computational grid."),
        ('is_adaptive_grid', 'bool', '0.1',
             "Default is False. Set true if grid resolution changes during execution."),
        ('thickness_level_1','float', '1.1',
             "Thickness of first surface ocean level (in meters)")
        ],
    }

# --------------------------------------------------------------------
# SUB-PROCESS: TUNING APPLIED: Any tuning used to optimise the parameters in this realm
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for ocean component',
    'properties': [
        ('description', 'str', '1.1',
             "General overview description of tuning: explain and motivate the main targets and metrics retained. &"
             "Document the relative weight given to climate performance metrics versus process oriented metrics, &"
             "and on the possible conflicts with parameterization level tuning. In particular describe any struggle &"
             "with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('global_mean_metrics_used', 'str', '0.N',
             "List set of metrics of the global mean state used in tuning model/component"),
        ('regional_metrics_used', 'str', '0.N',
             "List of regional metrics of mean state (e.g THC, AABW, regional means etc) used in tuning model/component"),
        ('trend_metrics_used', 'str', '0.N',
             "List observed trend metrics used in tuning model/component"),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: CONSERVATION PROPERTIES: Details of methodology needed to conserve variables between processes
# --------------------------------------------------------------------
DETAILS['conservation'] = {
    'description': 'Conservation in the ocean component',
    'properties': [
        ('description', 'str', '1.1', 'Brief description of conservation methodology'),
        ('scheme', 'ENUM:conservation_props_types', '1.N',
            'Properties conserved in the ocean by the numerical schemes'),
        ('consistency_properties', 'str','0.1',
            'Any additional consistency properties (energy conversion, pressure gradient discretisation, ...)?'),
        ('corrected_conserved_prognostic_variables', 'str', '0.1', # Can we constrains these variable
             "Set of variables which are conserved by *more* than the numerical scheme alone."),
        ('was_flux_correction_used', 'bool', '0.1',
             "Does conservation involved flux correction ?"),
        ],
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_family_types'] = {
    'description': 'Types of ocean models',
    'is_open': True,
    'members': [
        ('OGCM', None),
        ('slab ocean', None),
        ('mixed layer ocean', None)
        ]
    }

ENUMERATIONS['ocean_basic_approx_types'] = {
    'description': 'Types of basic approximation in ocean',
    'is_open': True,
    'members': [
        ('Primitive equations', None),
        ('Non-hydrostatic', None),
        ('Boussinesq', None)
        ]
    }

ENUMERATIONS['prognostic_vars_types'] = {
    'description': 'List of prognostic variables in ocean',
    'is_open': True,
    'members': [
        ('Potential temperature', None),
        ('Conservative temperature', None),
        ('Salinity', None),
        ('U-velocity', None),
        ('V-velocity', None),
        ('W-velocity', None),
        ('SSH', 'Sea Surface Height')
        ]
    }

ENUMERATIONS['seawater_eos_types'] = {
    'description': 'Types of seawater Equation of State in ocean',
    'is_open': True,
    'members': [
        ('Linear', None),
        ('Mc Dougall et al.', None),
        ('Jackett et al. 2006', None),
        ('TEOS 2010', None)
        ]
    }

ENUMERATIONS['seawater_eos_func_temp'] = {
    'description': 'Types of temperature used in EOS in ocean',
    'is_open': False,
    'members': [
        ('Potential temperature', None),
        ('Conservative temperature', None),
        ]
    }

ENUMERATIONS['seawater_eos_func_salt'] = {
    'description': 'Types of salinity used in EOS in ocean',
    'is_open': False,
    'members': [
        ('Practical salinity Sp', None),
        ('Absolute salinity Sa', None),
        ]
    }

ENUMERATIONS['seawater_eos_func_depth'] = {
    'description': 'Types of depth used in EOS in ocean',
    'is_open': False,
    'members': [
        ('Pressure (dbars)', None),
        ('Depth (meters)', None),
        ]
    }

ENUMERATIONS['seawater_freezing_point'] = {
    'description': 'Types of seawater freezing point equation in ocean',
    'is_open': True,
    'members': [
        ('TEOS 2010', None)
        ]
    }

ENUMERATIONS['bathymetry_ref_dates'] = {
    'description': 'List of reference dates for bathymetry in ocean',
    'is_open': True,
    'members': [
        ('Present day', None),
        ('21000 years BP', None),
        ('6000 years BP', None),
        ('LGM', 'Last Glacial Maximum'),
        ('Pliocene', None)
        ]
    }

ENUMERATIONS['conservation_props_types'] = {
    'description': 'List of properties that can be conserved in ocean',
    'is_open': True,
    'members': [
        ('Energy', None),
        ('Enstrophy', None),
        ('Salt', None),
        ('Volume of ocean', None),
        ('Momentum', None)
        ]
    }



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean Timestepping Framework'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel:timestepping_attributes'] = {
    'description': 'Properties of time stepping in ocean',
    'properties': [
        ('time_step', 'int', '1.1',
            'Ocean time step in seconds'),
        ('diurnal_cycle', 'ENUM:diurnal_cycle_types', '1.1',
            'Diurnal cycle type'),
        ]
    }

# This is the compact notation used of only one property
DETAILS['toplevel:timestepping_tracers_scheme'] = {
    'description': 'Properties of tracers time stepping in ocean',
    'properties': [
        ('tracers', 'ENUM:ocean_timestepping_types', '1.1',
            'Time stepping tracer scheme')
        ]
    }

DETAILS['toplevel:barotropic_solver_scheme'] = {
    'description': 'Barotropic solver in ocean',
    'properties': [
        ('barotropic_solver', 'ENUM:ocean_timestepping_types', '1.1',
            'Barotropic solver scheme'),
        ('type', 'ENUM:barotropic_solver_types', '1.1',
            'Barotropic solver type')
        ]
    }

DETAILS['toplevel:barotropic_momentum_scheme'] = {
    'description': 'Barotropic momentum solver in ocean',
    'properties': [
        ('barotropic_momentum', 'ENUM:ocean_timestepping_types', '1.1',
            'Barotropic momentum scheme')
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['diurnal_cycle_types'] = {
    'description': 'Types of diurnal cycle resolution in ocean',
    'is_open': True,
    'members': [
        ('None', 'No diurnal cycle in ocean'),
        ('Via coupling', 'Diurnal cycle via coupling frequency'),
        ('Specific treatment', 'Specific treament'),
        ]
    }

ENUMERATIONS['ocean_timestepping_types'] = {
    'description': 'Type of timestepping scheme in ocean',
    'is_open': True,
    'members': [
        ('Leap-frog + Asselin filter', 'Leap-frog scheme with Asselin filter'),
        ('Leap-frog + Periodic Euler backward solver', 'Leap-frog scheme with Periodic Euler backward solver'),
        ('Predictor-corrector', 'Predictor-corrector scheme'),
        ('AM3-LF (ROMS)', 'AM3-LF used in ROMS'),
        ('Forward-backward', 'Forward-backward scheme'),
        ('Forward operator', 'Forward operator scheme')
        ]
    }

ENUMERATIONS['barotropic_solver_types'] = {
    'description': 'Type of barotropic solver in ocean',
    'is_open': True,
    'members': [
        ('Preconditioned conjugate gradient', None),
        ('Sub cyling', None),
        ]
    }



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean lateral physics'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level lateral physics properties',
    'properties': [
        ('scheme', 'ENUM:latphys_transient_eddy_types', '1.1',
            'Type of transient eddy representation in ocean'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Momentum.
# --------------------------------------------------------------------
DETAILS['momentum'] = {
    'description': 'Properties of lateral physics for momentum in ocean',
    'properties': [],
    'detail_sets': [
        'operator',
        'eddy_viscosity_coeff'
        ]
    }

DETAILS['momentum:operator'] = {
    'description': 'Properties of lateral physics operator for momentum in ocean',
    'properties': [
        ('direction', 'ENUM:latphys_operator_direc_types', '1.1',
            'Direction of lateral physics momemtum scheme in the ocean'),
        ('order', 'ENUM:latphys_operator_order_types', '1.1',
            'Order of lateral physics momemtum scheme in the ocean'),
        ('discretisation', 'ENUM:latphys_operator_discret_types', '1.1',
            'Discretisation of lateral physics momemtum scheme in the ocean'),
        ]
    }

DETAILS['momentum:eddy_viscosity_coeff'] = {
    'description': 'Properties of eddy viscosity coeff in lateral physics momemtum scheme in the ocean',
    'properties': [
        ('type', 'ENUM:latphys_eddy_visc_coeff_types', '1.1',
            'Lateral physics momemtum eddy viscosity coeff type in the ocean'),
        ('constant_coefficient', 'int', '0.1',
            'If constant, value of eddy viscosity coeff in lateral physics momemtum scheme (in m2/s)'),
        ('variable_coefficient', 'str', '0.1',
            'If space-varying, describe variations of eddy viscosity coeff in lateral physics momemtum scheme'),
        ('coeff_background', 'int', '1.1',
            'Background value of eddy viscosity coeff in lateral physics momemtum scheme (in m2/s)'),
        ('coeff_backscatter', 'bool', '1.1',
            'Is there backscatter in eddy viscosity coeff in lateral physics momemtum scheme ?')
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Tracers.
# --------------------------------------------------------------------
DETAILS['tracers'] = {
    'description': 'Properties of lateral physics for tracers in ocean',
    'properties': [
        ('mesoscale_closure', 'bool', '1.1',
            'Is there a mesoscale closure in the lateral physics tracers scheme ?'),
        ('submesoscale_mixing', 'bool', '1.1',
            'Is there a submesoscale mixing parameterisation (i.e Fox-Kemper) in the lateral physics tracers scheme ?')

    ],
    'detail_sets': [
        'operator',
        'eddy_viscosity_coeff',
        'eddy_induced_velocity'
        ]
    }

DETAILS['tracers:operator'] = {
    'description': 'Properties of lateral physics operator for tracers in ocean',
    'properties': [
        ('direction', 'ENUM:latphys_operator_direc_types', '1.1',
            'Direction of lateral physics tracers scheme in the ocean'),
        ('order', 'ENUM:latphys_operator_order_types', '1.1',
            'Order of lateral physics tracers scheme in the ocean'),
        ('discretisation', 'ENUM:latphys_operator_discret_types', '1.1',
            'Discretisation of lateral physics tracers scheme in the ocean'),
        ]
    }

DETAILS['tracers:eddy_viscosity_coeff'] = {
    'description': 'Properties of eddy viscosity coeff in lateral physics tracers scheme in the ocean',
    'properties': [
        ('type', 'ENUM:latphys_eddy_visc_coeff_types', '1.1',
            'Lateral physics tracers eddy viscosity coeff type in the ocean'),
        ('constant_coefficient', 'int', '0.1',
            'If constant, value of eddy viscosity coeff in lateral physics tracers scheme (in m2/s)'),
        ('variable_coefficient', 'str', '0.1',
            'If space-varying, describe variations of eddy viscosity coeff in lateral physics tracers scheme'),
        ('coeff_background', 'int', '1.1',
            'Background value of eddy viscosity coeff in lateral physics tracers scheme (in m2/s)'),
        ('coeff_backscatter', 'bool', '1.1',
            'Is there backscatter in eddy viscosity coeff in lateral physics tracers scheme ?')
        ]
}

DETAILS['tracers:eddy_induced_velocity'] = {
    'description': 'Properties of eddy induced velocity (EIV) in lateral physics tracers scheme in the ocean',
    'properties': [
        ('type', 'ENUM:latphys_eiv_types', '1.1',
            'Type of EIV in lateral physics tracers in the ocean'),
        ('constant_val', 'int', '0.1',
            'If EIV scheme for tracers is constant, specify coefficient value (M2/s)'),
        ('flux_type', 'str', '1.1',
            'Type of EIV flux (advective or skew)'),
        ('added_diffusivity', 'str', '1.1',
            'Type of EIV added diffusivity (constant, flow dependent or none)')
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['latphys_transient_eddy_types'] = {
    'description': 'Type of transient eddy representation in ocean',
    'is_open': False,
    'members': [
        ('None', 'No transient eddies in ocean'),
        ('Eddy active', 'Full resolution of eddies'),
        ('Eddy admitting', 'Some eddy activity permitted by resolution'),
        ]
    }

ENUMERATIONS['latphys_operator_direc_types'] = {
    'description':'Type of lateral physics direction in ocean',
    'is_open': True,
    'members':[
        ('Horizontal', None),
        ('Isopycnal', None),
        ('Isoneutral', None),
        ('Geopotential', None),
        ('Iso-level', None)
        ]
    }

ENUMERATIONS['latphys_operator_order_types'] = {
    'description':'Type of lateral physics order in ocean',
    'is_open': True,
    'members':[
        ('Harmonic', 'Second order'),
        ('Bi-harmonic', 'Fourth order')
        ]
    }

ENUMERATIONS['latphys_operator_discret_types'] = {
    'description':'Type of lateral physics discretisation in ocean',
    'is_open': True,
    'members':[
        ('Second order', 'Second order'),
        ('Higher order', 'Higher order'),
        ('Flux limiter', None)
        ]
    }

ENUMERATIONS['latphys_eddy_visc_coeff_types'] = {
    'description':'Type of lateral physics eddy viscosity coeff in ocean',
    'is_open': True,
    'members':[
        ('Constant', None),
        ('Space varying', None),
        ('Time + space varying (Smagorinsky)', None)
        ]
    }

ENUMERATIONS['latphys_eiv_types'] = {
    'description':'Type of lateral physics eddy induced velocity in ocean',
    'is_open': True,
    'members':[
        ('GM', 'Gent & McWilliams')
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.utils.logger.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package logging utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import arrow

from pyesdoc.cv import constants



# Set of logging levels.
LOG_LEVEL_DEBUG = 'DUBUG'
LOG_LEVEL_INFO = 'INFO'
LOG_LEVEL_WARNING = 'WARNING'
LOG_LEVEL_ERROR = 'ERROR'
LOG_LEVEL_CRITICAL = 'CRITICAL'
LOG_LEVEL_FATAL = 'FATAL'

# Defaults.
_DEFAULT_APP = "PYESSV"
_DEFAULT_MODULE = "**"
_DEFAULT_INSTITUTE = "IPSL"

# Text to display when passed a null message.
_NULL_MSG = "-------------------------------------------------------------------------------"


def _get_formatted_message(msg, module, level, app, institute):
    """Returns a message formatted for logging.

    """
    if msg is None:
        return _NULL_MSG
    else:
        return "{0} [{1}] :: {2} {3} > {4} : {5}".format(
            unicode(arrow.get())[0:-6],
            level,
            institute,
            app,
            module,
            unicode(msg).strip()
            )


def log(
    msg=None,
    module=_DEFAULT_MODULE,
    level=LOG_LEVEL_INFO,
    app=_DEFAULT_APP,
    institute=_DEFAULT_INSTITUTE
    ):
    """Outputs a message to log.

    :param str msg: Message to be written to log.
    :param str module: Module emitting log message (e.g. MQ).
    :param str level: Message level (e.g. INFO).
    :param str app: Application emitting log message (e.g. libIGCM).
    :param str institute: Institute emitting log message (e.g. libIGCM).

    """
    # TODO use structlog.
    print _get_formatted_message(msg, module, level, app, institute)


def log_error(
    err,
    module=_DEFAULT_MODULE,
    app=_DEFAULT_APP,
    institute=_DEFAULT_INSTITUTE
    ):
    """Logs a runtime error.

    :param HermesClientException err: Error to be written to log.
    :param str module: Module emitting log message (e.g. MQ).
    :param str level: Message level (e.g. INFO).
    :param str app: Application emitting log message (e.g. libIGCM).
    :param str institute: Institute emitting log message (e.g. libIGCM).

    """
    msg = "!! {0} RUNTIME ERROR !! :: ".format(module)
    if issubclass(BaseException, err.__class__):
        msg += "{} :: ".format(err.__class__)
    msg += "{}".format(err)
    log(msg, module, LOG_LEVEL_ERROR, app, institute)



# -*- coding: utf-8 -*-

"""
.. module:: extender.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes functions for extending documents.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc import exceptions
from pyesdoc._extensions import cim
from pyesdoc._extensions import default
from pyesdoc.ontologies import type_info
from pyesdoc.utils import runtime



# Supported extenders keyed by document type.
SUPPORTED = {}
for o in {cim.v1, cim.v2}:
    for doc_type, doc_extender in o.SUPPORTED.iteritems():
        SUPPORTED[doc_type] = doc_extender


class _ExtensionContextInfo(object):
    """Document extension context information.

    """
    def __init__(self, doc, meta, ext):
        """Instance constructor.

        """
        self.doc = doc
        self.meta = meta
        self.ext = ext


class DocumentExtensionInfo(object):
    """Document extension information injected by extenders.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.children = []
        self.css_class = unicode()
        self.description = unicode()
        self.display_name = unicode()
        self.full_display_name = unicode()
        self.summary_fields = []
        self.type = unicode()
        self.viewer_url = unicode()


def extend(doc):
    """Invokes a document extender.

    :param doc: pyesdoc document instance.
    :type doc: object

    :returns: Extended document.
    :rtype: object

    """
    # Escape if extending null documents or already extended.
    if doc is None or hasattr(doc, 'ext'):
        return

    # Verify that document type is supported.
    if type(doc) not in type_info.TYPES:
        raise TypeError("Unsupported document type: {0}.".format(type(doc)))

    # Initialize document extension information.
    doc.ext = DocumentExtensionInfo()

    # Instantiate extension context.
    ctx = _ExtensionContextInfo(doc, doc.meta, doc.ext)

    # Step 1: invoke default pre-extenders.
    for extender in default.PRE_EXTENDERS:
        extender(ctx)

    # Step 2: invoke type specific extenders.
    if is_extendable(doc):
        mod = SUPPORTED[doc.type_key.lower()]
        # Step 2.1  Unpack child documents.
        try:
            mod.unpack_children(ctx)
        except AttributeError:
            pass

        # Step 2.2 Extend child documents (i.e. bottom up recursive extension).
        for child in ctx.ext.children:
            extend(child)

        # Step 2.3 Invoke extension callbacks.
        for extender in mod.get_extenders():
            try:
                extender(ctx)
            except Exception as err:
                raise exceptions.ExtendingException(err, extender)

    # Step 3: invoke default post-extenders.
    for extender in default.POST_EXTENDERS:
        extender(ctx)

    return doc


def is_extendable(doc):
    """Returns flag indicating whether a document is extensible or not.

    :param doc: pyesdoc document instance or type key.
    :type doc: object | str

    :returns: A flag indicating whether a document is extensible or not.
    :rtype: boolean

    """
    if not isinstance(doc, str):
        runtime.assert_doc(doc)

    doc_type = doc if isinstance(doc, str) else doc.type_key
    doc_type = doc_type.lower()

    return doc_type in SUPPORTED



# -*- coding: utf-8 -*-

"""
.. module:: constants.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# All files filter.
FILE_FILTER_ALL = "*.*"

# Error file filter.
FILE_FILTER_ERROR = "*.ERROR"

# Accepted file filter.
FILE_FILTER_INGESTED = "*_*_*_*.*"

# Set of all file filters.
FILE_FILTER_SET = {
	FILE_FILTER_ALL,
	FILE_FILTER_ERROR,
	FILE_FILTER_INGESTED
}




# -*- coding: utf-8 -*-

"""
.. module:: json.decoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Decodes a document from a JSON text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc._codecs.dictionary import decoder as dict_decoder
from pyesdoc.utils import convert



def decode(as_json):
    """Decodes a document from a UTF-8 encoded json text blob.

    :param as_json: Document json representation.
    :type as_json: unicode | str

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    # Convert to unicode.
    as_json = convert.str_to_unicode(as_json)

    # Convert to dictionary.
    as_dict = convert.json_to_dict(as_json)

    # Decode from dictionary.
    return dict_decoder.decode(as_dict)



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.typeset_for_misc_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v1.misc package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class DocumentSet(object):
    """A concrete class within the cim v1 type system.

    Represents a bundled set of documents.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DocumentSet, self).__init__()

        self.data = []                                    # data.DataObject (0.N)
        self.ensembles = []                               # activity.Ensemble (0.N)
        self.experiment = None                            # activity.NumericalExperiment (0.1)
        self.grids = []                                   # grids.GridSpec (0.N)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.model = None                                 # software.ModelComponent (0.1)
        self.platform = None                              # shared.Platform (0.1)
        self.simulation = None                            # activity.SimulationRun (0.1)





"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean biogeochemistry tracers'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level properties of tracers in ocean biogeochemistry',
    'properties': [
        ('sulfur_cycle', 'bool', '1.1',
            'Is sulfur cycle modeled ?'),
        ('nutrients', 'ENUM:nutrients_species', '1.N',
            'List nutrient species present in ocean biogeochemistry model'),
        ('nitrous_species', 'ENUM:nitrous_species', '0.N',
            'If nitrogen present, list nitrous species.'),
        ('nitrous_processes', 'ENUM:nitrous_processes', '0.N',
            'If nitrogen present, list nitrous processes.'),

        ]
    }

# --------------------------------------------------------------------
# PROCESS: detailed properties
# --------------------------------------------------------------------

DETAILS['Ecosystem'] = {
    'description': 'Ecosystem properties in ocean biogeochemistry',
    'properties': [
        ('upper_trophic_levels', 'str', '1.1',
            'Definition of upper trophic level (e.g. based on size) ?'),
        ('upper_trophic_levels_treatment', 'str', '1.1',
            'Define how upper trophic level are treated'),
        ]
    }

DETAILS['Ecosystem:phytoplancton'] = {
    'description': 'Phytoplancton properties in ocean biogeochemistry',
    'properties': [
        ('type', 'ENUM:plancton_types', '1.1',
            'Type of phytoplancton'),
        ('list', 'ENUM:phytoplancton_list', '0.N',
            'List species of phytoplancton if applicable'),
        ]
    }

DETAILS['Ecosystem:zooplancton'] = {
    'description': 'Zooplancton properties in ocean biogeochemistry',
    'properties': [
        ('type', 'ENUM:plancton_types', '1.1',
            'Type of zooplancton'),
        ('list', 'ENUM:zooplancton_list', '0.N',
            'List species of zooplancton if applicable'),
        ]
    }

DETAILS['Disolved_organic_matter'] = {
    'description': 'Disolved organic matter properties in ocean biogeochemistry',
    'properties': [
        ('bacteria_representation', 'bool', '1.1',
            'Is there bacteria representation ?'),
        ('lability', 'ENUM:lability_treatment', '1.1',
            'Describe treatment of lability in dissolved organic matter'),
        ]
    }


DETAILS['Particules'] = {
    'description': 'Particulate carbon properties in ocean biogeochemistry',
    'properties': [
        ('method', 'ENUM:particules_method', '1.1',
            'How is particulate carbon represented in ocean biogeochemistry?'),
        ('types', 'ENUM:prognostic_particules_types', '0.1',
            'If prognostic, types of particulate matter taken into account'),
        ('size', 'ENUM:prognostic_particules_sizes', '0.1',
            'If prognostic, describe if a particule size spectrum is used to represent distribution of particules in water volume'),
        ('discrete_size', 'str', '0.1',
            'If prognostic and discrete size, describe which size classes are used'),
        ('sinking_speed', 'ENUM:prognostic_particules_sinking_speed', '0.1',
            'If prognostic, method for calculation of sinking speed of particules'),
        ]
    }
#
# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['nutrients_species'] = {
    'description': 'Nutrients species in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Nitrogen (N)', None),
        ('Phosphorous (P)', None),
        ('Silicium (S)', None),
        ('Iron (Fe)', None),
        ]
    }

ENUMERATIONS['nitrous_species'] = {
    'description': 'Nitrous species in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Nitrates (NO3)', None),
        ('Amonium (NH4)', None),
       ]
    }

ENUMERATIONS['nitrous_processes'] = {
    'description': 'Nitrous processes in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Dentrification', None),
        ('N fixation', None),
       ]
    }

ENUMERATIONS['plancton_types'] = {
    'description': 'Plancton types in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('None', None),
        ('Generic', None),
        ('List (specify below)', None),       ]
    }


ENUMERATIONS['phytoplancton_list'] = {
    'description': 'Phytoplancton list in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('diatoms', None),
        ('Nfixers', None),
        ('calcifiers', None),
        ('picophytoplancton', None),
       ]
    }

ENUMERATIONS['zooplancton_list'] = {
    'description': 'Zooplancton list in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('microzooplancton', None),
        ('mesozooplancton', None),
       ]
    }

ENUMERATIONS['lability_treatment'] = {
    'description': 'Lability treatment of dissolved organic matter in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('none', None),
        ('labile', 'Less than a few days'),
        ('semi-labile', 'Few days to a few years'),
        ('refractory', 'Over a few years'),
        ]
    }

ENUMERATIONS['particules_method'] = {
    'description': 'Particulate carbon representation in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Diagnostic', None),
        ('Diagnostic (Martin profile)', None),
        ('Diagnostic (Balast)', None),
        ('Prognostic', None),
       ]
    }

ENUMERATIONS['prognostic_particules_types'] = {
    'description': 'Prognostic particulate carbon types in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('POC', None),
        ('PIC (calcite)', None),
        ('PIC (aragonite', None),
        ('BSi', None),
       ]
    }

ENUMERATIONS['prognostic_particules_sizes'] = {
    'description': 'Prognostic particulate carbon size in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('Full size spectrum', None),
        ('Discrete size classes (specify which below)', None),
        ('No size spectrum used', None),
       ]
    }

ENUMERATIONS['prognostic_particules_sinking_speed'] = {
    'description': 'Prognostic particulate carbon sinking speed in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Constant', None),
        ('Function of particule size', None),
        ('Function of particul type (balast)', None),
       ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.definition.table_attribute.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps a table row attribute defined within dreq2Defn.xml.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.drq import utils



# Map of table xml attribute names to value covertors.
_CONVERTORS = {
    'required': lambda v: True if v == "True" else False,
}

# Map of table row attribute type to python type.
_TYPE_MAP = {
    "xs:string": str,
    "xs:integer": int,
    "xs:float": float,
    "xs:boolean": bool,
    "aa:st__integerList": list,
    "aa:st__integerListMonInc": list,
    "aa:st__stringList": list,
    "aa:st__floatList": list,
    "aa:st__uid": str,
}

# Map of table attribute type to default value.
_DEFAULT_VALUE_MAP = {
    "aa:st__integerList": list(),
    "aa:st__integerListMonInc": list(),
    "aa:st__stringList": list(),
    "aa:st__floatList": list()
}


class TableAttribute(object):
    """Wraps a table attribute defined within dreq2Defn.xml.

    """
    def __init__(self, table, elem):
        """Instance constructor.

        :param dreq.definition.table.Table table: Associated definition table.
        :param xml.etree.Element elem: XML element declared within data request definition.

        """
        # Initialize from XML.
        utils.init_from_xml(self, elem, elem.keys(), _CONVERTORS)

        # Switch to pythonic labels.
        self.label_drq = self.label
        self.label = utils.get_label(self.label)
        self._sort_key = self.label.lower()

        # Map drq types.
        self.default_value = _DEFAULT_VALUE_MAP.get(self.type)
        self.type_python = _TYPE_MAP[self.type]

        # Format use_class field.
        try:
            self.use_class
        except AttributeError:
            self.use_class = None
        else:
            if self.use_class == "":
                self.use_class == None


    def __repr__(self):
        """Instance representation.

        """
        return self.label


    @property
    def is_internal_link(self):
        """Gets flag indicating whether the attribute is an internal link or not.

        """
        return self.use_class == "internalLink"





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.utils.__init__.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Utility functions sub-package.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.ontology
   :platform: Unix, Windows
   :synopsis: Represents an ontology definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from operator import add
from functools import reduce

from pyesdoc.mp.ontologies.core.class_ import Class
from pyesdoc.mp import utils



class Ontology(object):
    """Represents an ontology, i.e. a set of classes organised into packages.

    """
    def __init__(self, name, version, doc_string, packages):
        """Instance constructor.

        :param str name: Ontology name.
        :param str version: Ontology version.
        :param str doc_string: Ontology documentation string.
        :param iterable packages: Set of associated packages.

        """
        self.classes = reduce(add, [p.classes for p in packages])
        self.decodings = reduce(add, [p.decodings for p in packages])
        self.doc_string = doc_string
        self.entities = []
        self.enums = reduce(add, [p.enums for p in packages])
        self.enum_members = reduce(add, [e.members for e in self.enums])
        self.name = name
        self.op_name = None
        self.op_version = None
        self.packages = sorted(packages, key=lambda p: p.name)
        self.properties = reduce(add, [c.properties for c in self.classes])
        self.property_types = [p.type for p in self.properties]
        self.sub_classed = tuple()
        self.types = sorted(self.classes + self.enums)
        self.version = version

        # Set derived information.
        for setter in [
            _set_relations,
            _set_base_classes,
            _set_sub_classes,
            _set_entities,
            _set_property_type_info,
            _set_class_imports,
            _set_circular_imports,
            _set_associated_packages,
            _set_associated_packages_for_import,
            _set_package_external_type_refs,
            _set_collection_sort_orders,
        ]:
            setter(self)


    def __repr__(self):
        """Instance string representation.

        """
        return "{0} v{1}".format(self.name, self.version)


    def get_type(self, name):
        """Returns type with matching name.

        :param str name: Fully qualified name of target type.

        """
        pkg_name = name.split('.')[0]
        type_name = name.split('.')[1]
        for t in self.types:
            if t.package.name == pkg_name and t.name == type_name:
                return t


def _set_relations(ontology):
    """Sets relations between various nodes within an ontology.

    """
    for pkg in ontology.packages:
        pkg.ontology = ontology
        for cls in pkg.classes:
            cls.package = pkg
            for prp in cls.properties + cls.computed_properties + cls.constraints:
                prp.package = pkg
                prp.cls = cls
        for enum in pkg.enums:
            enum.package = pkg
            for enum_member in enum.members:
                enum_member.enum = enum


def _set_base_classes(ontology):
    """Sets base classes.

    """
    for cls in [c for c in ontology.classes if c.base]:
        base_cls = ontology.get_type(cls.base)
        if base_cls:
            cls.base = base_cls
        else:
            msg = "Base class not found :: class = {0}.{1} :: base = {2}"
            msg = msg.format(cls.package, cls, cls.base)
            utils.log(msg)


def _set_sub_classes(ontology):
    """Sets sub classes.

    """
    # Identify classes with a base class as these are potential targets.
    with_base = [c for c in ontology.classes if c.base]

    # Set sub-classed classes.
    sub_classed = {}
    for i in ontology.classes:
        sub_classed[i] = [(c.bases.index(i), c) for c in with_base if i in c.bases]
    sub_classed = {k: v for k, v in sub_classed.items() if v}

    # Set ontology & package sub-class sets.
    ontology.sub_classed = sorted(sub_classed.keys(), key=lambda i: str(i))
    for pkg in ontology.packages:
        pkg.sub_classed = [c for c in ontology.sub_classed if c.package == pkg]

    # Set sub-class hierachies.
    for i in sub_classed:
        depth = max([sc[0] for sc in sub_classed[i]]) + 1
        hierachy = {k: sorted(sc[1] for sc in sub_classed[i] if sc[0] == k) for k in range(depth)}
        i.sub_class_hierachy = [(level, hierachy[level]) for level in hierachy]
        i.sub_classes = [sc[1] for sc in sorted(sub_classed[i], key=lambda i: str(i))]


def _set_entities(ontology):
    """Sets package entities (i..e classes with a meta attribute).

    """
    for pkg in ontology.packages:
        pkg.entities = sorted([c for c in pkg.classes if c.is_entity])
        ontology.entities += pkg.entities


def _set_property_type_info(ontology):
    """Sets a property type information.

    """
    for pt in ontology.property_types:
        pt.cls = pt.package = None
        if pt.is_complex:
            type_ = ontology.get_type(pt.name)
            if type_ is not None:
                pt.is_class = isinstance(type_, Class)
                pt.ontology = ontology
                pt.package = type_.package
                pt.cls = type_


def _set_class_imports(ontology):
    """Assigns set of intra-class imports.

    """
    def _append(cls, pkg, type_):
        """Appends an import to the classes set of imports.

        """
        if pkg != cls.package.name or type_ != cls.name:
            cls.imports.append((pkg, type_))

    for cls in ontology.classes:
        if cls.base is not None:
            _append(cls, cls.base.package.name, cls.base.name)
        for p in [p for p in cls.properties if p.type.is_complex]:
            _append(cls, p.type.name_of_package, p.type.name_of_type)


def _set_circular_imports(ontology):
    """Assigns set of intra-class circular imports.

    """
    for c in ontology.classes:
        c_import = (c.package.name, c.name)
        for prp in [p for p in c.properties if p.type.is_class]:
            p_type = ontology.get_type(prp.type.name)
            if c_import in p_type.imports:
                p_type.imports.remove(c_import)
                p_type.circular_imports.append(c_import)


def _set_associated_packages(ontology):
    """Assigns set of intra-package associations.

    """
    for pkg in ontology.packages:
        pkg.associated += [c.base.package for c in pkg.classes if c.base and c.base.package != pkg]
        pkg.associated += [prp.type.package for prp in pkg.properties if prp.type.package and prp.type.package != pkg]
        pkg.associated = list(set(pkg.associated))


def _set_associated_packages_for_import(ontology):
    """Assigns set of intra-package associations required for import.

    """
    for pkg in ontology.packages:
        pkg.associated_for_import += [c.base.package for c in pkg.classes if c.base and c.base.package != pkg]
        pkg.associated_for_import += [prp.type.package for prp in pkg.properties if prp.type.package and prp.type.package != pkg and prp.name == "meta"]
        pkg.associated_for_import = list(set(pkg.associated_for_import))


def _set_package_external_type_refs(ontology):
    """Assigns set of a package's external type references so that .

    """
    for pkg in ontology.packages:
        for prp in pkg.properties:
            if prp.type.is_complex and \
               prp.type.name_of_package != pkg.name:
                pkg.external_types.append(prp.type)


def _set_collection_sort_orders(ontology):
    """Sets sort order of various collections.

    """
    ontology.classes = sorted(ontology.classes, key=str)
    ontology.enums = sorted(ontology.enums, key=str)
    ontology.types = sorted(ontology.types, key=str)
    ontology.enum_members = sorted(ontology.enum_members, key=str)
    ontology.entities = sorted(ontology.entities, key=str)
    ontology.properties = sorted(ontology.properties, key=str)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.property
   :platform: Unix, Windows
   :synopsis: Represents an ontological type property definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.core.type_ import Type



class Property(object):
    """Represents a property within an ontology.

    """
    def __init__(self, name, type_name, cardinality, doc_string):
        """Instance constructor.

        :param str name: Property name.
        :param str doc_string: Property documentation string.
        :param str type_name: Property type name.
        :param str cardinality: Type of relationship to associated class (i.e. 0.1 | 1.1 | 0.N | 1.N).

        """
        if name.lower() in ("ext", ):
            raise AttributeError("{0} is a reserved property name.".format(name))

        self.cls = None
        self.decodings = []
        self.doc_string = doc_string if doc_string is not None else u""
        self.cardinality = cardinality
        self.is_collection = cardinality.split('.')[1] == 'N'
        self.is_linked_document = (name == 'meta')
        self.is_required = cardinality.split('.')[0] != '0'
        self.max_occurs = cardinality.split('.')[1]
        self.min_occurs = cardinality.split('.')[0]
        self.name = name
        self.type = Type(type_name)


    def __repr__(self):
        """Instance string representation.

        """
        return self.name




# -*- coding: utf-8 -*-

"""
.. module:: html.encoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Encodes a document as an HTML text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import collections
import inspect
import os

import tornado.template as template

import pyesdoc
from pyesdoc import exceptions
from pyesdoc._codecs.html import fieldsets
from pyesdoc.ontologies import cim
from pyesdoc.utils import runtime



# Path to template directory.
_DIR_TEMPLATES = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")

# Document set template.
_DOCUMENT_SET_TEMPLATE = "core/document_set.html"

# Mapping of doucment type to template.
_TEMPLATE_TYPE_MAPPINGS = {
    # CIM v1 templates.
    cim.v1.activity.Ensemble: "cim_1/activity_ensemble.html",
    cim.v1.activity.NumericalExperiment: "cim_1/activity_numerical_experiment.html",
    cim.v1.activity.SimulationRun: "cim_1/activity_simulation_run.html",
    cim.v1.data.DataObject: "cim_1/data_data_object.html",
    cim.v1.grids.GridSpec: "cim_1/grids_grid_spec.html",
    cim.v1.shared.Platform: "cim_1/shared_platform.html",
    cim.v1.software.ModelComponent: "cim_1/software_model_component.html",
    cim.v1.quality.CimQuality: "cim_1/quality_cimquality.html",
    # CIM v2 templates.
    cim.v2.designing.NumericalExperiment: "cim_2/designing_numerical_experiment.html",
    cim.v2.designing.Project: "cim_2/designing_project.html",
    cim.v2.shared.Citation: "cim_2/shared_citation.html",
    cim.v2.shared.Party: "cim_2/shared_party.html"
}


def encode(doc):
    """Encodes a document to HTML.

    :param object doc: Document being encoded.

    :returns: An HTML representation of a document.
    :rtype: str

    """
    # Convert to iterable.
    try:
        iter(doc)
    except TypeError:
        document_set = [doc]
    else:
        document_set = doc

    # Filter out fragments & non-templated documents.
    document_set = [d for d in document_set if hasattr(d, "meta") and \
                                               isinstance(d, tuple(_TEMPLATE_TYPE_MAPPINGS.keys()))]

    # Escape if no documents.
    if not document_set:
        return

    # Sort.
    document_set = sorted(document_set, key=lambda d: d.meta.sort_key)

    # Extend.
    for document in document_set:
        pyesdoc.extend(document)

    # JIT initialize templates.
    if isinstance(_DOCUMENT_SET_TEMPLATE, str):
        _init_templates()

    # Return generated template.
    try:
        return _DOCUMENT_SET_TEMPLATE.generate(
            document_set=document_set,
            document_group_set=_get_group_set(document_set),
            generate_document=_generate,
            pyesdoc=pyesdoc
            )
    except Exception as error:
        raise exceptions.EncodingException('html', doc, error)


def _init_templates():
    """Initializes templates.

    """
    global _DOCUMENT_SET_TEMPLATE

    loader = template.Loader(_DIR_TEMPLATES)
    _DOCUMENT_SET_TEMPLATE = loader.load(_DOCUMENT_SET_TEMPLATE)
    for doc_type, template_path in _TEMPLATE_TYPE_MAPPINGS.items():
        _TEMPLATE_TYPE_MAPPINGS[doc_type] = loader.load(template_path)


def _log(msg):
    """Emits a template logging message - used for debugging.

    """
    print "TEMPLATE:", msg


class TemplateInfo(object):
    """Template processing information wrapper.

    """
    def __init__(self,
                 data,
                 fieldset=None,
                 fieldset_type=None):
        self.field = None
        self.fieldset = fieldsets.create(fieldset)
        if inspect.isfunction(self.fieldset):
            self.fieldset = self.fieldset(data)
        self.fieldset_type = fieldset_type or "namevalue"
        try:
            iter(data)
        except TypeError:
            self.data = data
            self.dataset = [data]
        else:
            self.data = self.dataset = data
        self.dataset = [i for i in self.dataset if i is not None]


def _generate(document):
    """Returns generated document.

    """
    try:
        return _TEMPLATE_TYPE_MAPPINGS[type(document)].generate(
            doc=document,
            TemplateInfo=TemplateInfo,
            pyesdoc=pyesdoc,
            load=pyesdoc.archive.load_references,
            cim=pyesdoc.ontologies.cim,
            log=_log,
            collections=collections
            )
    except Exception as err:
        runtime.log_error("Template generation error: {0}".format(err.message))
        raise err


def _get_group_set(document_set):
    """Returns grouped document set.

    """
    def get_sort_key(doc):
        """Returns key used for document sorting.

        """
        return doc.ext.full_display_name.lower()

    def get_group_key(doc):
        """Returns key used for document grouping.

        """
        return "{0}-{1}".format(doc.meta.type_sort_key,
                                doc.meta.type_display_name)

    # Initialize group sets.
    group_set = collections.defaultdict(list)
    for document in document_set:
        group_key = get_group_key(document)
        group_set[group_key].append(document)

    # Sort group sets.
    for group_key, document_set in group_set.iteritems():
        group_set[group_key] = sorted(document_set, key=get_sort_key)

    return group_set



# -*- coding: utf-8 -*-

"""
.. module:: json.encoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encodes a term to JSON.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.cv.codecs.dictionary import encoder as dict_encoder
from pyesdoc.cv.utils import convert



def encode(instance):
    """Encodes an instance of a domain model class as a JSON text blob.

    :param pyesdoc.cv.Entity instance: A domain model class instance to be encoded as a JSON text blob.

    :returns: Instance encoded as a JSON text blob.
    :rtype: unicode

    """
    # Convert to dictionary.
    as_dict = dict_encoder.encode(instance)

    # Return JSON.
    return convert.dict_to_json(as_dict)



# -*- coding: utf-8 -*-

"""
.. module:: shared_enums_doc.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 shared package document related enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def doc_relationship_direction_type():
    """Creates and returns instance of relationship_direction_type enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('toTarget', None),
            ('fromTarget', None),
        ],
    }


def doc_relationship_type():
    """Creates and returns instance of document_relationship_type enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('similarTo', None),
            ('other', None),
            ('laterVersionOf', None),
            ('previousVersionOf', None),
            ('fixedVersionOf', None),
        ],
    }


def doc_status_type():
    """Status of cim document.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('complete', None),
            ('incomplete', None),
            ('in-progress', None),
        ],
    }


def doc_type():
    """Creates and returns instance of doc_type enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('downscalingSimulation', None),
            ('statisticalModelComponent', None),
            ('simulationRun', None),
            ('assimilation', None),
            ('simulationComposite', None),
            ('numericalExperiment', None),
            ('dataProcessing', None),
            ('ensemble', None),
            ('dataObject', None),
            ('gridSpec', None),
            ('cimQuality', None),
            ('platform', None),
            ('processorComponent', None),
            ('modelComponent', None),
        ],
    }






# -*- coding: utf-8 -*-

"""
.. module:: test_config.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Tests data request configuration wrapper.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import dreq


def test_is_loaded():
    """dreq-tests: config: verify config is loaded into memory.

    """
    assert isinstance(dreq.config, dreq.ConfigWrapper)
    assert len(dreq.config) == len(dreq.config.tables) == 24



"""
.. module:: pyesdoc.mp.ontologies.generators.python.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package initializer.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.schemas import cim
from pyesdoc.mp.ontologies.schemas.utils import get_schema
from pyesdoc.mp.ontologies.core.schema_validation import is_valid
from pyesdoc.mp.ontologies.core.schema_validation import validate



"""
.. module:: pyesdoc.mp.generate.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: pyesdoc.mp code generator.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import argparse

import pyesdoc



# Define command line arguments.
_parser = argparse.ArgumentParser("ES-DOC Code Generator.")
_parser.add_argument(
    "-s", "--schema-name",
    help="Target ontology schema. [default = cim]",
    dest="schema_name",
    type=str
    )
_parser.add_argument(
    "-v", "--schema-version",
    help="Target ontology schema version. [default = 1]",
    dest="schema_version",
    type=str
    )
_parser.add_argument(
    "-l", "--language",
    help="Target programming language. [default = python]",
    dest="language",
    type=str
    )
_parser.add_argument(
    "-o", "--output-dir",
    help="Path to a directory to which generated code will be written",
    dest="output_dir",
    type=str
    )


# Set command line options.
args = _parser.parse_args()

# Set ontology schema.
schema = pyesdoc.mp.get_schema(args.schema_name, args.schema_version)
if schema is None:
    raise pyesdoc.mp.exceptions.UnsupportedOntologySchema('Unsupported schema: {0} v{1}.'.format(args.schema_name, args.schema_version))

# Generate.
pyesdoc.mp.generate(schema, args.language, args.output_dir)



"""A realm grid sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Ruth Petrie'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Ruth Petrie'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Sea Ice grid'

# --------------------------------------------------------------------
# DISCRETISATION: Description of the horizontal discretization and numerics.
# --------------------------------------------------------------------
DETAILS['discretisation'] = {
    'description': 'Sea ice discretisation',
    'properties': [],
    'detail_sets': [
        'horizontal',
        'vertical',
        ]
    }

DETAILS['discretisation:horizontal'] = {
    'description': 'Sea ice discretisation in the horizontal',
    'properties': [
        ('grid', 'ENUM:horizontal_grid', '1.1',
            "Grid on which sea ice is horizontal discretised?"),
        ('grid_type', 'ENUM:grid_structure', '1.1',
            "What is the type of sea ice grid?"),
        ('scheme', 'ENUM:numerical_scheme', '1.1',
            "What is the advection scheme?"),
        ('thermodynamics_time_step', 'int', '1.1',
            "What is the time step in the sea ice model thermodynamic component in seconds."),
        ('dynamics_time_step', 'int', '1.1',
            "What is the time step in the sea ice model dynamic component in seconds."),
        ('additional_details', 'str', '0.1',
            "Specify any additional horizontal discretisation details.")
        ]
    }

DETAILS['discretisation:vertical'] = {
    'description': 'Sea ice vertical properties',
    'properties': [
        ('layering', 'ENUM:vertical_layering', '1.N',
            "What type of sea ice vertical layers are implemented for purposes of thermodynamic calculations?"),
        ('number_of_layers', 'int', '1.1',
            "If using multi-layers specify how many."),
        ('additional_details', 'str', '0.1',
            "Specify any additional vertical grid details.")
        ]
    }



# --------------------------------------------------------------------
# SEAICE-CATEGROIES: Description of sea ice categories.
# --------------------------------------------------------------------
DETAILS['seaice_categories'] = {
    'description': 'What method is used to represent sea ice categories ?',
    'properties': [
        ('has_mulitple_categories', 'bool', '1.1',
            "Set to true if the sea ice model has multiple sea ice categories."),
        ('number_of_categories', 'int', '1.1',
            "If using sea ice categories specify how many."),
        ('category_limits', 'str', '1.1',
            "If using sea ice categories specify each of the category limits."),
        ('ice_thickness_distribution_scheme', 'str', '1.1',
            "Describe the sea ice thickness distribution scheme"),
        ('other', 'str', '0.1',
            "If the sea ice model does not use sea ice categories specify any additional details."),
        ]
    }

# --------------------------------------------------------------------
# Snow on ice grid information
# --------------------------------------------------------------------
DETAILS['snow_on_seaice'] = {
    'description': 'Snow on sea ice details',
    'properties': [
        ('has_snow_on_ice', 'bool', '1.1',
            "Is snow on ice represented in this model?"),
        ('number_of_snow_levels', 'int', '1.1',
            "Number of vertical levels of snow on ice?"),
        ('snow_fraction', 'str', '1.1',
            "Describe how the snow fraction on sea ice is determined"),
        ('additional_details', 'str', '0.1',
            "Specify any additional details related to snow on ice.")
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['horizontal_grid'] = {
    'description': 'Grid which is the sea ice horizontally discretised on?',
    'is_open': True,
    'members': [
        ('Ocean grid',
            'Sea ice is horizontally discretised on the ocean grid'),
        ('Atmosphere Grid',
            'Sea ice is horizontally discretised on the atmospheric grid'),
        ('Own Grid',
            'Sea ice is horizontally discretised on its own independent grid'),
        ]
    }

ENUMERATIONS['vertical_layering'] = {
    'description': 'Sea ice layering types',
    'is_open': True,
    'members': [
        ('Zero-layer',
            'Simulation has no internal ice thermodynamics.'),
        ('Two-levels',
            'Simulation uses two layers.'),
        ('Multi-level',
            'Simulation uses more than two layers'),
       ]
    }

ENUMERATIONS['numerical_scheme'] = {
    'description': 'Numerical schemes',
    'is_open': True,
    'members': [
        ('Finite differences', None),
        ('Finite elements', None),
        ('Finite volumes',  None),
       ]
    }

ENUMERATIONS['grid_structure'] = {
    'description': 'Grid structures',
    'is_open': True,
    'members': [
        ('Structured grid', None),
        ('Unstructured grid', None),
        ('Adaptive grid', 'Computational grid changes during the run'),
    ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: xml.encoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Encodes a document as an XML text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import datetime
import uuid
import xml.etree.ElementTree as ET

from pyesdoc._codecs.dictionary import encoder as dict_encoder
from pyesdoc.utils import convert



# Default XML document encoding.
_UNICODE = "utf-8"

# Mappings between collection name suffixes and collection item suffixes.
_suffix_maps = (
    ('ies', 'y'),
    ('List', ''),
    ('s', ''),
)


def _get_item_tag(tag):
    """Returns tag name of a collection item.

    """
    for suffix, replacement in _suffix_maps:
        if tag.endswith(suffix):
            return tag[0:len(tag) - len(suffix)] + replacement

    return tag + 'Item'


def _is_encodable_simple(val):
    """Determines whether a simple value requires encoding.

    """
    return val not in (None, str(), int())


def _encode_simple(xml, val):
    """Encodes a simple value.

    """
    # Format according to type.
    if val in (None, 'None'):
        return u''
    elif isinstance(val, datetime.datetime):
        val = val.isoformat().replace('T', ' ')
    elif isinstance(val, datetime.date):
        val = val.isoformat()
    elif isinstance(val, datetime.time):
        val = val.isoformat()
    else:
        val = convert.str_to_unicode(val)
    if val is None or len(val) == 0:
        val = u''

    # Assign to xml.
    xml.text = val.strip()


def _encode_collection(xml, tag, collection):
    """Encodes a collection.

    """
    tag = _get_item_tag(tag)
    for item in collection:
        if isinstance(item, dict):
            _encode_complex(ET.SubElement(xml, tag), item)
        else:
            _encode_simple(ET.SubElement(xml, tag), item)


def _encode_complex(xml, d):
    """Encodes a complex type.

    """
    # Iterate dictionary keys sorted in alphabetical order.
    for k in sorted(d.iterkeys()):
        # Get attribute value.
        v = d[k]

        # Process according to value type.
        # ... iterable
        if isinstance(v, list):
            if len(v):
                elem = ET.SubElement(xml, k)
                _encode_collection(elem, k, v)

        # ... complex
        elif isinstance(v, dict):
            if len(v.keys()):
                elem = ET.SubElement(xml, k)
                _encode_complex(elem, v)

        # ... simple
        elif _is_encodable_simple(v):
            # ... type key is assigned as an attribute
            if k in ('ontologyTypeKey',):
                xml.attrib[k] = v
            else:
                elem = ET.SubElement(xml, k)
                _encode_simple(elem, v)

    return xml


def _get_xml_tag(doc):
    """Returns document root xml tag.

    """
    tag = type(doc).type_key.split('.')[3]
    tag = convert.str_to_camel_case(tag)

    return tag


def encode(doc):
    """Encodes a document.

    :param doc: Document being encoded.
    :type doc: object

    :returns: An encoded document representation.
    :rtype: unicode

    """
    # Encode as a dictionary.
    as_dict = dict_encoder.encode(doc)

    # Format dictionary keys.
    as_dict = convert.dict_keys(as_dict, convert.str_to_camel_case)

    # Encode to an etree xml element.
    tag = _get_xml_tag(doc)
    xml = _encode_complex(ET.Element(tag), as_dict)

    return ET.tostring(xml, _UNICODE)



# Cardinality constraint.
CONSTRAINT_TYPE_CARDINALITY = 'cardinality'

# Constant constraint.
CONSTRAINT_TYPE_CONSTANT = 'constant'

# Type constraint.
CONSTRAINT_TYPE_TYPEOF = 'type'

# Regular expression constraint.
CONSTRAINT_TYPE_REGEX = 'regex'

# set of supported constraint types.
CONSTRAINT_TYPES = {
	CONSTRAINT_TYPE_CARDINALITY,
	CONSTRAINT_TYPE_CONSTANT,
	CONSTRAINT_TYPE_REGEX,
	CONSTRAINT_TYPE_TYPEOF
}



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.cv.__init__.py

   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL / CeCILL
   :platform: Unix
   :synopsis: Python earth science standard vocabulary package intializer.

.. moduleauthor:: IPSL (ES-DOC) <dev@esdocumentation.org>

"""
from pyesdoc.cv import archive
from pyesdoc.cv.io_mgr import read_authority
from pyesdoc.cv.io_mgr import write_authority
from pyesdoc.cv.codecs import from_dict
from pyesdoc.cv.codecs import from_json
from pyesdoc.cv.codecs import to_dict
from pyesdoc.cv.codecs import to_json
from pyesdoc.cv.constants import *
from pyesdoc.cv.exceptions import *
from pyesdoc.cv.factory import create_authority
from pyesdoc.cv.factory import create_scope
from pyesdoc.cv.factory import create_collection
from pyesdoc.cv.factory import create_term
from pyesdoc.cv.model import Term
from pyesdoc.cv.model import Authority
from pyesdoc.cv.model import Collection
from pyesdoc.cv.model import Scope





# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.typeset_for_quality_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v1.quality package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class CimQuality(object):
    """A concrete class within the cim v1 type system.

    The starting point for a quality record.  It can contain any number of issues and reports.  An issue is an open-ended description of some issue about a CIM instance.  A record is a prescribed description of some specific quantitative measure that has been applied to a CIM instance.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(CimQuality, self).__init__()

        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.reports = []                                 # quality.Report (0.N)


class Evaluation(object):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of evaluation class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Evaluation, self).__init__()

        self.date = None                                  # datetime.datetime (0.1)
        self.description = None                           # unicode (0.1)
        self.did_pass = None                              # bool (0.1)
        self.explanation = None                           # unicode (0.1)
        self.specification = None                         # unicode (0.1)
        self.specification_hyperlink = None               # unicode (0.1)
        self.title = None                                 # unicode (0.1)
        self.type = None                                  # unicode (0.1)
        self.type_hyperlink = None                        # unicode (0.1)


class Measure(object):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of measure class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Measure, self).__init__()

        self.description = None                           # unicode (0.1)
        self.identification = None                        # unicode (0.1)
        self.name = None                                  # unicode (0.1)


class Report(object):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of report class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Report, self).__init__()

        self.date = None                                  # datetime.datetime (0.1)
        self.evaluation = None                            # quality.Evaluation (1.1)
        self.evaluator = None                             # shared.ResponsibleParty (0.1)
        self.measure = None                               # quality.Measure (1.1)


class CimFeatureType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of cim_feature_type enum.
    """
    is_open = False
    members = [
        "file",
        "diagnostic"
        ]
    descriptions = [
        "None",
        "None"
        ]


class CimResultType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of cim_result_type enum.
    """
    is_open = False
    members = [
        "plot",
        "document",
        "logfile"
        ]
    descriptions = [
        "None",
        "None",
        "None"
        ]


class CimScopeCodeType(object):
    """An enumeration within the cim v1 type system.

    This would cover quality issues with the CIM itself.
    """
    is_open = False
    members = [
        "dataset",
        "software",
        "service",
        "model",
        "modelComponent",
        "simulation",
        "experiment",
        "numericalRequirement",
        "ensemble",
        "file"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None"
        ]


class QualityIssueType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of quality_issue_type enum.
    """
    is_open = False
    members = [
        "metadata",
        "data_format",
        "data_content",
        "data_indexing",
        "science"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None",
        "None"
        ]


class QualitySeverityType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of quality_severity_type enum.
    """
    is_open = False
    members = [
        "cosmetic",
        "minor",
        "major"
        ]
    descriptions = [
        "None",
        "None",
        "None"
        ]


class QualityStatusType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of quality_status_type enum.
    """
    is_open = False
    members = [
        "reported",
        "confirmed",
        "partially_resolved",
        "resolved"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: test_content_section_item.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Tests data request content section items.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import dreq



def test_type():
    """dreq-tests: content-section-item: verify section item types.

    """
    for section in dreq.content:
        for item in section:
            assert isinstance(item, dreq.ContentSectionItem)


def test_required():
    """dreq-tests: content-section-item: verify required section items.

    """
    for section in dreq.content:
        for item in section:
        	for attribute in section.cfg.required_attributes:
    			assert hasattr(item, attribute.name)


def test_value_type():
    """dreq-tests: content-section-item: verify value type.

    """
    for section in dreq.content:
        for item in section:
            for attribute in section.cfg.scalar_attributes:
                if hasattr(item, attribute.name):
                    assert type(getattr(item, attribute.name)) == attribute.type_python



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.content.section_item.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps a section item defined within dreq.xml.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.drq import utils
from pyesdoc.drq.definition.table_attribute import TableAttribute



class SectionItem(object):
    """Wraps a section item defined within dreq.xml.

    """
    def __init__(self, table, elem):
        """Instance constructor.

        :param dreq.definition.Table table: Associated definition table.
        :param xml.etree.ElementTree elem: Section item xml info.

        """
        self._TABLE = table
        utils.init_from_xml(
            self,
            elem,
            sorted([i.label for i in table]),
            {i.label: i.type_python for i in table if i.type_python != list}
            )
        self._sort_key = self.label.lower()
        self.links = SectionItemLinks(self)


    def __repr__(self):
        """Instance representation.

        """
        return self.label or self.uid or str(type(self))


    def __len__(self):
        """Returns number of item attributes.

        """
        return len(self._TABLE.attributes)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter([(i, self[i]) for i in self._TABLE])


    def __getitem__(self, key):
        """Returns a child table attribute.

        """
        attr = None
        if isinstance(key, TableAttribute):
            attr = key
        elif isinstance(key, int):
            attr = self._TABLE[key]
        else:
            key = str(key).strip().lower()
            for attribute in self._TABLE:
                if attribute.label.lower() == key:
                    attr = attribute
                    break
        if attr is not None:
            return getattr(self, attr.label)


class SectionItemLinks(object):
    """Wraps the collection of linked data request sections.

    """
    def __init__(self, item):
        """Instance constructor.

        :param dreq.content.SectionItem item: Associated section item.

        """
        for attr in [i for i in item._TABLE if i.is_internal_link]:
            short_name, long_name = utils.get_link_labels(attr.label)
            if short_name is not None:
                setattr(self, short_name, None)
            if short_name != long_name and long_name is not None:
                setattr(self, long_name, None)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.codecs.__init__.py
   :copyright: Copyright "Sep 4, 2013", Earth System Documentation
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates transformations of terms from one format to another.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.cv.codecs import dictionary
from pyesdoc.cv.codecs import json
from pyesdoc.cv.constants import ENCODING_DICT
from pyesdoc.cv.constants import ENCODING_JSON
from pyesdoc.cv.model import TYPES



# Codecs mapped by encoding.
_CODECS = {
	ENCODING_DICT: dictionary,
	ENCODING_JSON: json
}

# Map of encodings to allowed input types when decoding.
_DECODE_TYPE_WHITELIST = {
    ENCODING_DICT : (dict, ),
    ENCODING_JSON : (str, unicode)
}


def _decode(target, encoding=ENCODING_JSON):
    """Returns a decoded domain model class instance.

    :param basestring|dict target: A domain model class instance representation.
    :param str encoding: A supported encoding (dict|json).

    :returns: A domain model class instance.
    :rtype: object

    """
    if target is None:
        raise ValueError("Cannot decode a null pointer.")
    if not encoding in _CODECS:
        raise NotImplementedError('Unsupported term encoding :: {0}.'.format(encoding))
    if not isinstance(target, _DECODE_TYPE_WHITELIST[encoding]):
        err = "Representation unsupported: must be one of {}."
        err = err.format(_DECODE_TYPE_WHITELIST[encoding])
        raise TypeError(err)

    return _CODECS[encoding].decode(target)


def _encode(target, encoding=ENCODING_JSON):
    """Returns an encoded domain model class instance|collection.

    :param object|list target: Domain model class instance|collection.
    :param str encoding: A supported encoding (dict|json).

    :returns: Target encoded accordingly.
    :rtype: unicode|dict|list

    """
    if target is None:
        raise ValueError("Null encoding error")
    if encoding not in _CODECS:
        raise NotImplementedError('Invalid encoding: {}'.format(encoding))

    if isinstance(target, TYPES):
        encoded = _CODECS[encoding].encode(target)
        if isinstance(encoded, basestring):
            encoded = encoded.strip()
        return encoded

    else:
        return [_encode(d, encoding) for d in target]


def from_json(representation):
    return _decode(representation, ENCODING_JSON)

def from_dict(representation):
    return _decode(representation, ENCODING_DICT)

def to_json(instance):
    return _encode(instance, ENCODING_JSON)

def to_dict(instance):
    return _encode(instance, ENCODING_DICT)



# -*- coding: utf-8 -*-

"""
.. module:: test_general.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc general tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import inspect

import nose.tools

import pyesdoc
import test_utils as tu
import test_types as tt



# Test ontology constants.
_CIM = 'cim'
_CIM_V1 = '1'
_CIM_PACKAGE = 'software'
_CIM_TYPE = 'modelComponent'


def _test_module_setup(mod):
    """Test that the test document module is correctly setup."""
    tu.assert_bool(mod in tt.MODULES, True)
    tu.assert_bool(mod in tt.INITIAL_STATE, True)
    for field in tt.STATE_FIELDS:
        tu.assert_bool(hasattr(mod, field), True)


def _test_version():
    """Test package version."""
    tu.assert_str(pyesdoc.__version__, "0.9.2.1")


def _test_module_reset(mod):
    """Test that the test document modules are correctly reset."""
    # Assert module state is cached.
    assert mod in tt.INITIAL_STATE

    # Update state.
    for field in tt.STATE_FIELDS:
        setattr(mod, field, 'XXX')
        tu.assert_str(getattr(mod, field), 'XXX')

    # Reset.
    tt.reset(mod)

    # Assert initial state.
    for field in tt.STATE_FIELDS:
        state = tt.INITIAL_STATE[mod][field]
        assert getattr(mod, field) == state


def _test_module_file_open(mod):
    """Test opening module test files."""
    assert tu.get_test_file(mod.DOC_FILE) is not None


def _test_import(mod):
    """Test module import."""
    assert inspect.ismodule(mod)


def test_imports_01():
    """Test importing packages - 1."""
    for mod in (
        pyesdoc,
        pyesdoc.archive,
        pyesdoc.archive.config,
        pyesdoc.constants,
        pyesdoc.factory,
        pyesdoc.io,
        pyesdoc.ontologies,
        pyesdoc.ontologies.cim.v1,
        pyesdoc.ontologies.cim.v2,
        pyesdoc.options,
        pyesdoc.extensions,
        pyesdoc.extensions.default,
        pyesdoc.extensions.extender,
        pyesdoc.publishing,
        pyesdoc.serialization,
        pyesdoc.utils,
        pyesdoc.utils.convert,
        pyesdoc.utils.runtime,
        pyesdoc.validation,
        pyesdoc.validation.graph,
        pyesdoc.validation.validator,
        ):
        tu.init(_test_import, "import module", mod)
        yield _test_import, mod


def test_imports_02():
    """Test importing packages - 2."""
    cim = pyesdoc.ontologies.cim
    for mod in (
        cim,
        cim.v1,
        cim.v1.decoder,
        cim.v1.decoder_for_activity_package,
        cim.v1.decoder_for_data_package,
        cim.v1.decoder_for_grids_package,
        cim.v1.decoder_for_misc_package,
        cim.v1.decoder_for_quality_package,
        cim.v1.decoder_for_shared_package,
        cim.v1.decoder_for_software_package,
        cim.v1.typeset,
        cim.v1.typeset_for_activity_package,
        cim.v1.typeset_for_data_package,
        cim.v1.typeset_for_grids_package,
        cim.v1.typeset_for_misc_package,
        cim.v1.typeset_for_quality_package,
        cim.v1.typeset_for_shared_package,
        cim.v1.typeset_for_software_package,
        cim.v1.typeset_meta,
        ):
        tu.init(_test_import, "import cim module", mod)
        yield _test_import, mod


def _test_is_supported_ontology():
    """Test supported ontologies."""
    # supported
    assert pyesdoc.is_supported(_CIM, _CIM_V1)

    # unsupported
    assert not pyesdoc.is_supported('x', _CIM_V1)
    assert not pyesdoc.is_supported(_CIM, 'x')


def _test_is_supported_type_01():
    """Test supported ontology types - positive."""
    # supported
    assert pyesdoc.is_supported(_CIM, _CIM_V1, _CIM_PACKAGE, _CIM_TYPE)
    for ontology, version, package, typeof in pyesdoc.list_types():
        assert pyesdoc.is_supported(ontology, version, package, typeof)


def _test_is_supported_type_02():
    """Test supported ontology types - negative."""
    # unsupported
    assert not pyesdoc.is_supported('x', _CIM_V1, _CIM_PACKAGE, _CIM_TYPE)
    assert not pyesdoc.is_supported(_CIM, 'x', _CIM_PACKAGE, _CIM_TYPE)
    assert not pyesdoc.is_supported(_CIM, _CIM_V1, 'x', _CIM_TYPE)
    assert not pyesdoc.is_supported(_CIM, _CIM_V1, _CIM_PACKAGE, 'x')


def _test_list_types():
    """Test listing supported types."""
    # supported - all
    types = pyesdoc.list_types()
    tu.assert_int(len(types), 177)

    # supported - cim v1
    types = pyesdoc.list_types(_CIM, _CIM_V1)
    tu.assert_int(len(types), 103)

    # unsupported
    types = pyesdoc.list_types('x', 'x')
    tu.assert_int(len(types), 0)


def _test_set_option_01():
    """Test setting package options - positive."""
    api_url = 'https://es-doc.org'
    api_url_old = pyesdoc.get_option('api_url')
    pyesdoc.set_option('api_url', api_url)
    tu.assert_str(api_url, pyesdoc.get_option('api_url'))
    pyesdoc.set_option('api_url', api_url_old)


@nose.tools.raises(pyesdoc.InvalidOptionException)
def _test_set_option_02():
    """Test setting package options - negative.

    """
    pyesdoc.set_option('xxx', 'xxx')


def test():
    """Runs set of general unit tests."""
    for mod in tt.MODULES:
        tu.init(_test_module_file_open, 'open document module test file', mod)
        yield _test_module_file_open, mod

    for mod in tt.MODULES:
        tu.init(_test_module_reset, 'document module reset', mod)
        yield _test_module_reset, mod

    for mod in tt.MODULES:
        tu.init(_test_module_setup, 'document module setup', mod)
        yield _test_module_setup, mod

    for func in (
        _test_version,
        _test_set_option_01,
        _test_set_option_02,
        _test_is_supported_ontology,
        _test_is_supported_type_01,
        _test_is_supported_type_02,
        _test_list_types,
        ):
        tu.init(func, func.__doc__[5:])
        yield func




"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Characteristics of the dynamical core'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "General dynamical core properties",
    'properties': [
        ('timestepping_type', 'ENUM:timestepping_type', '1.1',
            'Timestepping framework type'),
        ('prognostic_variables', 'ENUM:prognostic_variables', '1.N',
            'List of the model prognostic variables'),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: SUB-PROCESSES
# --------------------------------------------------------------------
DETAILS['top_boundary'] = {
    'description': 'Type of boundary layer at the top of the model',
    'properties': [
        ('top_boundary_condition', 'ENUM:top_boundary_condition', '1.1',
            'Top boundary condition'),
        ('top_heat', 'str', '1.1',
            'Top boundary heat treatment'),
        ('top_wind', 'str', '1.1',
            'Top boundary wind treatment'),
        ]
    }

DETAILS['lateral_boundary'] = {
    'description': 'Type of lateral boundary condition (if the model is a regional model)',
    'properties': [
        ('condition', 'ENUM:lateral_boundary', '0.1',
             'Type of lateral boundary condition'),
        ]
    }

DETAILS['diffusion_horizontal'] = {
    'description': 'Horizontal diffusion scheme',
    'properties': [
        ('scheme_name', 'str', '1.1',
             'Horizontal diffusion scheme name'),
        ('scheme_method', 'ENUM:diffusion_horizontal_scheme_method', '1.1',
             'Horizontal diffusion scheme method'),
        ]
    }

DETAILS['advection_tracers'] = {
    'description': 'Tracer advection scheme',
    'properties': [
        ('scheme_name', 'ENUM:advection_tracers_scheme_name', '1.1',
            'Tracer advection scheme name'),
        ('scheme_characteristics', 'ENUM:advection_tracers_scheme_characteristics', '1.N',
            'Tracer advection scheme characteristics'),
        ('conserved_quantities', 'ENUM:advection_tracers_conserved_quantities', '1.N',
            'Tracer advection scheme conserved quantities'),
        ('conservation_method', 'ENUM:advection_tracers_conservation_method', '1.1',
            'Tracer advection scheme conservation method'),
        ]
    }

DETAILS['advection_momentum'] = {
    'description': 'Momentum advection scheme',
    'properties': [
        ('scheme_name', 'ENUM:advection_momentum_scheme_name', '1.1',
            'Momentum advection schemes name'),
        ('scheme_characteristics', 'ENUM:advection_momentum_scheme_characteristics', '1.N',
         'Momentum advection scheme characteristics'),
        ('scheme_staggering_type', 'ENUM:advection_momentum_scheme_staggering_type', '1.1',
            'Momentum advection scheme staggering type'),
        ('conserved_quantities', 'ENUM:advection_momentum_conserved_quantities', '1.N',
            'Momentum advection scheme conserved quantities'),
        ('conservation_method', 'ENUM:advection_momentum_conservation_method', '1.1',
            'Momentum advection scheme conservation method'),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['timestepping_type'] = {
    'description': 'Type of time stepping scheme',
    'is_open': True,
    'members': [
        ('Adams-Bashforth', None),
        ('explicit', None),
        ('implicit', None),
        ('semi-implicit', None),
        ('leap frog', None),
        ('multi-step', None),
        ('Runge Kutta fifth order', None),
        ('Runge Kutta second order', None),
        ('Runge Kutta third order', None),
        ]
    }

ENUMERATIONS['top_boundary_condition'] = {
    'description': 'Type of boundary layer at the top of the model',
    'is_open': True,
    'members': [
        ('sponge layer', None),
        ('radiation boundary condition', None),
        ]
    }

ENUMERATIONS['lateral_boundary'] = {
    'description': 'Type of lateral boundary condition (if the model is a regional model)',
    'is_open': True,
    'members': [
        ('sponge layer', None),
        ('radiation boundary condition', None),
        ]
    }

ENUMERATIONS['prognostic_variables'] = {
    'description': 'List of the model prognostic variables',
    'is_open': True,
    'members': [
        ('surface pressure', None),
        ('wind components', None),
        ('divergence/curl', None),
        ('temperature', None),
        ('potential temperature', None),
        ('total water', None),
        ('water vapour', None),
        ('water liquid', None),
        ('water ice', None),
        ('total water moments', None),
        ('clouds', None),
        ('radiation', None),
        ]
    }

# Made this a string attribute: no obvious convention for consistent scheme naming convention
# ENUMERATIONS['diffusion_horizontal_scheme_name'] = {
#    'description': 'Commonly used name for the horizontal diffusion scheme',
#    'is_open': True,
#    'members': []
#    }

ENUMERATIONS['diffusion_horizontal_scheme_method'] = {
    'description': 'Numerical method used by the horizontal diffusion scheme',
    'is_open': True,
    'members': [
        ('iterated Laplacian', None),
        ('bi-harmonic', None)
        ]
    }

ENUMERATIONS['advection_tracers_scheme_name'] = {
    'description': 'Commonly used name for the tracer advection scheme',
    'is_open': True,
    'members': [
            ('Heun', None),
            ('Roe and VanLeer', None),
            ('Roe and Superbee', None),
            ('Prather', None),
            ('UTOPIA', None),
        ]
    }

ENUMERATIONS['advection_tracers_scheme_characteristics'] = {
    'description': 'Characteristics of the numerical scheme used for the advection of tracers',
    'is_open': True,
    'members': [
            ('Eulerian', None),
            ('modified Euler', None),
            ('Lagrangian', None),
            ('semi-Lagrangian', None),
            ('cubic semi-Lagrangian', None),
            ('quintic semi-Lagrangian', None),
            ('mass-conserving', None),
            ('finite volume', None),
            ('flux-corrected', None),
            ('linear', None),
            ('quadratic', None),
            ('quartic', None),
        ]
    }

# TODO: is this information nescessary?
ENUMERATIONS['advection_tracers_conserved_quantities'] = {
    'description': 'Quantities conserved through the tracers advection scheme',
    'is_open': True,
    'members': [
        ('dry mass', None)
        ]
    }

# TODO: is this information nescessary?
ENUMERATIONS['advection_tracers_conservation_method'] = {
    'description': 'Method used to ensure conservation in the tracers advection scheme',
    'is_open': True,
    'members': [
        ('conservation fixer', None),
        ('Priestley algorithm', None),
        ]
    }

ENUMERATIONS['advection_momentum_scheme_name'] = {
    'description': 'Commonly used name for the momentum advection scheme',
    'is_open': True,
    'members': [
        ('VanLeer', None),
        ('Janjic', None),
        ('SUPG (Streamline Upwind Petrov-Galerkin)', None),
        ]
    }

ENUMERATIONS['advection_momentum_scheme_characteristics'] = {
    'description': 'Characteristics of the numerical scheme used for the advection of momentum',
    'is_open': True,
    'members': [
        ('2nd order', None),
        ('4th order', None),
        ('cell-centred', None),
        ('staggered grid', None),
        ('semi-staggered grid', None),
        ]
    }

ENUMERATIONS['advection_momentum_scheme_staggering_type'] = {
    'description': 'If scheme characteristics specify staggered grid, describe the type of staggering',
    'is_open': True,
    'members': [
        ('Arakawa B-grid', None),
        ('Arakawa C-grid', None),
        ('Arakawa D-grid', None),
        ('Arakawa E-grid', None),
        ]
    }

ENUMERATIONS['advection_momentum_conserved_quantities'] = {
    'description': 'Quantities conserved through the tracers advection scheme',
    'is_open': True,
    'members': [
        ("Angular momentum", None),
        ("Horizontal momentum", None),
        ("Enstrophy", None),
        ("Mass", None),
        ("Total energy", None),
        ("Vorticity", None),
        ]
    }

# TODO: is this information nescessary?
ENUMERATIONS['advection_momentum_conservation_method'] = {
    'description': 'Method used to ensure conservation in the momentum advection scheme',
    'is_open': True,
    'members': [
            ('conservation fixer', None),
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.partitions.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of partitions being managed by client.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import inspect
import os

import pyesdoc.cv
from pyesdoc.cv import constants
from pyesdoc.cv._partition_cache import PartitionCache
from pyesdoc.cv._partition_io import PartitionIO



class Partition(object):
    """Encpasulates info related to a partition.

    """
    def __init__(self, domain, io_dir):
        """Instance constructor.

        """
        if len(domain.split(".")) != 2:
            raise ValueError("Partition domain is invalid.")
        if not os.path.isdir(io_dir):
            raise ValueError("Partition I/O directory is invalid")

        self.io = PartitionIO(domain, io_dir)
        self.cache = PartitionCache(self.io)
        self.domain = unicode(domain)

        # Pull in delegated functions.
        self.is_cached = self.cache.exists
        self.is_written = self.io.exists
        self.query = self.cache.query
        self.show = self.cache.show
        self.get_count = self.cache.get_count
        self.load = self.cache.init


    def create(self, kind, name):
        """Factory function that returns a newly created term.

        :param str kind: Term kind (e.g. model-componenet)
        :param str name: Term preferred name (e.g. Aerosols)

        :returns: Instantiated term.
        :rtype: pyesdoc.cv.Term

        """
        new_term = pyesdoc.cv.Term(self.domain, kind, name)
        new_term.idx = self.get_count(kind) + 1

        return new_term


    def save(self, target):
        """Saves a set of newly created terms to file system.

        """
        def _save(term, emit_log_entry=False):
            """Saves term to file system & updates local cache.

            """
            self.io.write(term)
            if not term.is_cached:
                self.cache.add(term, emit_log_entry)

        if isinstance(target, pyesdoc.cv.Term):
            _save(target, True)
        elif inspect.isfunction(target):
            for term in target(self):
                _save(term)
        elif inspect.isgenerator(target):
            for term in target:
                _save(term)
        else:
            raise ValueError("Cannot save term data")


    def _govern(self, term, status):
        """Sets governance state of a term.

        """
        if term.status != status:
            term.status = status
            self.save(term)

        return term


    def accept(self, term):
        """Marks a term as accepted.

        :param pyesdoc.cv.Term term: Term being marked as accepted.

        """
        return self._govern(term, constants.GOVERNANCE_STATUS_ACCEPTED)


    def reject(self, term):
        """Marks a term as rejected.

        :param pyesdoc.cv.Term term: Term being marked as rejected.

        """
        return self._govern(term, constants.GOVERNANCE_STATUS_REJECTED)


    def deprecate(self, term):
        """Marks a term as deprecated.

        :param pyesdoc.cv.Term term: Term being marked as deprecated.

        """
        return self._govern(term, constants.GOVERNANCE_STATUS_DEPRECATED)


    def destroy(self, term):
        """Removes a term from existence leaving no trace.

        :param pyesdoc.cv.Term term: Term to be destroyed.

        """
        self.io.delete(term)
        self.cache.remove(term)


    def parse(self, kind, name):
        """Parses a term name and returns an input Removes a term from existence leaving no trace.

        :param str kind: Term kind (e.g. model-componenet)
        :param str name: Term preferred name (e.g. Aerosols)

        """
        name = unicode(name).lower()
        for term in self.cache.query(kind):
            for term_name in term.all_names:
                if term_name.lower() == name:
                    return term.name


    def add_synonym(self, kind, name, new_synonym):
        """Appends a synonym to a term's set of synonyms.

        :param str kind: Term kind (e.g. model)
        :param str name: Term preferred name (e.g. ipsl-cm5a-lr)
        :param str new_synonym: A new synonym for the term (e.g. IPSL-LR)

        """
        term = self.cache.query(kind, name)
        if term is None:
            raise ValueError("Term could not be found: {} :: {}".format(kind, name))

        term.add_synonym(new_synonym)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.computed_property
   :platform: Unix, Windows
   :synopsis: Represents an ontological computed property definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class ClassConstraint(object):
    """Represents a class constraint defined within an ontology.

    """
    def __init__(self, property_name, typeof, value):
        """Instance constructor.

        :param str property_name: Property name.
        :param str typeof: Constraint type.
        :param str value: Constrained value.

        """
        if property_name.lower() in ("ext", ):
            raise AttributeError("{0} is a reserved property name.".format(property_name))

        self.cls = None
        self.property_name = property_name
        self.typeof = typeof
        self.value = value
        self.package = None


    def __repr__(self):
        """Instance string representation.

        """
        return "class constraint; {}::{}::{}".format(self.property_name, self.typeof, self.value)


    @property
    def key(self):
        """Returns a key for disambiguating amongst a set of constraints.

        """
        return "{}-{}".format(self.property_name, self.typeof)



# -*- coding: utf-8 -*-

"""
.. module:: cim_1.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: CIM v1 HTML field sets.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import collections

from pyesdoc._codecs.html.fieldsets.field import Field
from pyesdoc.ontologies import cim



_SEPARATOR = "  |  "

# Experimental relationship types and their labels.
_EXPERIMENTAL_RELATIONSHIP_LABELS = [
    ("is_initialized_by", "Parent"),
    ("is_sibling_of", "Siblings"),
    ("is_initializer_of", "Children"),
    ("is_perturbation_from", "Control"),
    ("is_control_for", "Provides control to"),
    ("is_constrained_by", "Boundary Conditions From"),
    ("is_constrainer_of", "Provides constraints to"),
]


def _get_related_experiments(e, r_type):
    result = []
    for i in sorted(e.related_experiments, key=lambda i: i.name):
        if i.relationship == r_type:
            result.append((i.name, i.viewer_url))

    return result


def _get_designing_project_experiments(p):
    """Returns a dynamically constructed fieldset.

    """
    fields = []

    # Governed experiments by tier.
    for tier in sorted(p.ext.tiered_experiments):
        experiments = p.ext.tiered_experiments[tier]
        description = "Tier {}".format(tier)
        fields.append(Field(description,
            link_factory=[(i.name, i.viewer_url) for i in experiments])
        )

    # Other experiments.
    if p.ext.other_experiments:
        fields.append(Field('Other',
            link_factory=lambda p: [(i.name, i.viewer_url) for i in p.ext.other_experiments])
            )

    return fields


def _get_designing_numericalexperiment_relationships(e):
    """Returns a dynamically constructed fieldset.

    """
    fields = []
    for typeof, label in _EXPERIMENTAL_RELATIONSHIP_LABELS:
        if [i for i in e.related_experiments if i.relationship == typeof]:
            fields.append(Field(label, link_factory=_get_related_experiments(e, typeof)))

    return fields


# Document field sets.
FIELDSETS = {
    'cim.2.designing.numericalexperiment-overview': [
        Field('MIP Era', path='meta.project'),
        Field('Related MIPs',
            link_factory=lambda v: [(i.name, i.viewer_url) for i in sorted(v.related_mips, key=lambda v: v.name)]
            ),
        Field('Institute', path='meta.institute'),
        Field('Canonical Name', path='canonical_name'),
        Field('Alternative Names', path='alternative_names',
            input_formatter=lambda v: _SEPARATOR.join(v),
            predicate=lambda v: len(v.alternative_names) > 0
            ),
        Field('Long Name', path='long_name'),
        Field('Tier', path='tier'),
        Field('Description', path='description'),
        Field('Rationale', path='rationale'),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(','))
            )
    ],

    'cim.2.designing.numericalexperiment-relationships':
        _get_designing_numericalexperiment_relationships,

    str(cim.v2.designing.EnsembleRequirement) : [
        Field('Name', path='canonical_name'),
        Field('Description', path='description'),
        Field('Ensemble Type', path='ensemble_type'),
        Field('Minimum Size', path='minimum_size'),
        Field('Is Conformance Requested ?', path='is_conformance_requested'),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(','))
            )
    ],

    str(cim.v2.designing.ForcingConstraint) : [
        Field('Name', path='canonical_name'),
        Field('Description', path='description'),
        Field('Rationale', path='rationale'),
        Field('Forcing Type', path='forcing_type', capitalize_value=True),
        Field('Is Conformance Requested ?', path='is_conformance_requested'),
        Field('Is Conformance Info Required ?', path='is_conformance_info_required'),
        Field('Scope', path='scope', capitalize_value=True),
        Field('Data Link', path='data_link.name',
            link_path=lambda v: v.data_link.availability[0].linkage,
            predicate=lambda v: v.data_link is not None and \
                                len(v.data_link.availability) > 0
            ),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(','))
            )
    ],

    str(cim.v2.designing.MultiEnsemble) : [
        Field('Name', path='canonical_name'),
        Field('Description', path='description'),
        Field('Forcing Type', path='forcing_type'),
        Field('Is Conformance Requested ?', path='is_conformance_requested'),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(','))
            )
    ],

    str(cim.v2.designing.NumericalRequirement) : [
        Field('Name', path='canonical_name'),
        Field('Description', path='description'),
        Field('Ensemble Type', path='ensemble_type'),
        Field('Minimum Size', path='minimum_size'),
        Field('Is Conformance Requested ?', path='is_conformance_requested'),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(','))
            )
    ],

    str(cim.v2.designing.TemporalConstraint) : [
        Field('Start Date', path='start_date.value'),
        Field('Required Duration', path='required_duration',
            input_formatter=lambda v: "{} {}".format(v.length, v.units)
            ),
        Field('Description', path='description'),
        Field('Is Conformance Requested ?', path='is_conformance_requested'),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(','))
            )
    ],

    'cim.2.designing.project-overview':[
        Field('Name', path='name'),
        Field('Long Name', path='long_name'),
        Field('Homepage', path='homepage'),
        Field('Description', path='description'),
        Field('Rationale', path='rationale'),
        Field('Objectives', path='objectives', predicate=lambda v: len(v.objectives) > 0),
        Field('Sub Projects',
            link_factory=lambda p: [(i.name, i.viewer_url) for i in p.sub_projects],
            predicate=lambda v: len(v.sub_projects) > 0
            ),
        Field('Keywords', path='keywords',
            input_formatter=lambda v: _SEPARATOR.join(v.split(',')))
        ],

    'cim.2.designing.project-experiments':
        _get_designing_project_experiments,

    'cim.2.shared.citation' : [
        Field('DOI', path='doi',
            link_path=lambda v: "https://doi.org/{}".format(v.doi)
            ),
        Field('Title', path='citation_detail', link_path="url.linkage"),
        Field('Context', path='context'),
        Field('Abstract', path='abstract')
    ],

    'cim.2.shared.party' : [
        Field('Name', path='name', link_path="url.linkage"),
        Field('Address', path='address'),
        Field('Email', path='email', email_path='email')
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: constants.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Notebook constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Set of valid QC states.
QC_STATES = {0, 1, 2}



# -*- coding: utf-8 -*-

"""
.. module:: test_content_section.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Tests data request content sections.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import dreq



# Set of section properties mapped from <section> xml attributes.
_SECTION_PROPERTIES = {
    'id',
    'label',
    'title',
    'uid',
    'use_class',
    }


def test_type():
    """dreq-tests: content-section: verify section types.

    """
    for section in dreq.content:
        assert isinstance(section, dreq.ContentSection)


def test_properties():
    """dreq-tests: content-section: verify section properties.

    """
    for section in dreq.content:
        for name in _SECTION_PROPERTIES:
            assert hasattr(section, name), name


def test_property_pythonic_label():
    """dreq-tests: content-section: verify each section has an injected label_pythonic property.

    """
    for section in dreq.content:
		try:
			assert section.label_pythonic == dreq.LABEL_MAP[section.label]
		except KeyError:
			assert section.label_pythonic == section.label


def test_direct_reference():
    """dreq-tests: content-section: verify direct access to specific section.

    """
    assert dreq.content.CMORvar is not None
    assert dreq.content.experiment is not None
    assert dreq.content.exptgroup is not None
    assert dreq.content.grids is not None
    assert dreq.content.mip is not None
    assert dreq.content.modelConfig is not None
    assert dreq.content.objective is not None
    assert dreq.content.objectiveLink is not None
    assert dreq.content.remarks is not None
    assert dreq.content.requestItem is not None
    assert dreq.content.requestLink is not None
    assert dreq.content.requestVar is not None
    assert dreq.content.requestVarGroup is not None
    assert dreq.content.spatialShape is not None
    assert dreq.content.standardname is not None
    assert dreq.content.structure is not None
    assert dreq.content.tableSection is not None
    assert dreq.content.temporalShape is not None
    assert dreq.content.timeSlice is not None
    assert dreq.content.var is not None
    assert dreq.content.varChoice is not None
    assert dreq.content.varChoiceLinkC is not None
    assert dreq.content.varChoiceLinkR is not None


def test_direct_reference_pythonic_label():
    """dreq-tests: content-section: verify access to specific sections via it's pythonic label.

    """
    assert dreq.content.cmor_variable == dreq.content.CMORvar
    assert dreq.content.experiment_group == dreq.content.exptgroup
    assert dreq.content.model_config == dreq.content.modelConfig
    assert dreq.content.objective_link == dreq.content.objectiveLink
    assert dreq.content.request_item == dreq.content.requestItem
    assert dreq.content.request_link == dreq.content.requestLink
    assert dreq.content.request_variable == dreq.content.requestVar
    assert dreq.content.request_variable_group == dreq.content.requestVarGroup
    assert dreq.content.spatial_shape == dreq.content.spatialShape
    assert dreq.content.standard_name == dreq.content.standardname
    assert dreq.content.table_section == dreq.content.tableSection
    assert dreq.content.temporal_shape == dreq.content.temporalShape
    assert dreq.content.time_slice == dreq.content.timeSlice
    assert dreq.content.var_choice == dreq.content.varChoice
    assert dreq.content.var_choice_link_c == dreq.content.varChoiceLinkC
    assert dreq.content.var_choice_link_r == dreq.content.varChoiceLinkR




# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.typeset.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Typeset information for the cim v1 ontology.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from typeset_for_activity_package import *
from typeset_for_data_package import *
from typeset_for_grids_package import *
from typeset_for_misc_package import *
from typeset_for_quality_package import *
from typeset_for_shared_package import *
from typeset_for_software_package import *


import typeset_for_activity_package as activity
import typeset_for_data_package as data
import typeset_for_grids_package as grids
import typeset_for_misc_package as misc
import typeset_for_quality_package as quality
import typeset_for_shared_package as shared
import typeset_for_software_package as software



# Module exports.
__all__ = [
    # Packages
    activity,
    data,
    grids,
    misc,
    quality,
    shared,
    software,
    # Classes
    Activity,
    BoundaryCondition,
    Conformance,
    DownscalingSimulation,
    Ensemble,
    EnsembleMember,
    Experiment,
    ExperimentRelationship,
    ExperimentRelationshipTarget,
    InitialCondition,
    LateralBoundaryCondition,
    MeasurementCampaign,
    NumericalActivity,
    NumericalExperiment,
    NumericalRequirement,
    NumericalRequirementOption,
    OutputRequirement,
    PhysicalModification,
    Simulation,
    SimulationComposite,
    SimulationRelationship,
    SimulationRelationshipTarget,
    SimulationRun,
    SpatioTemporalConstraint,
    DataContent,
    DataDistribution,
    DataExtent,
    DataExtentGeographical,
    DataExtentTemporal,
    DataExtentTimeInterval,
    DataHierarchyLevel,
    DataObject,
    DataProperty,
    DataRestriction,
    DataStorage,
    DataStorageDb,
    DataStorageFile,
    DataStorageIp,
    DataTopic,
    CoordinateList,
    GridExtent,
    GridMosaic,
    GridProperty,
    GridSpec,
    GridTile,
    GridTileResolutionType,
    SimpleGridGeometry,
    VerticalCoordinateList,
    DocumentSet,
    CimQuality,
    Evaluation,
    Measure,
    Report,
    Calendar,
    Change,
    ChangeProperty,
    Citation,
    ClosedDateRange,
    Compiler,
    Daily360,
    DataSource,
    DateRange,
    DocMetaInfo,
    DocReference,
    License,
    Machine,
    MachineCompilerUnit,
    OpenDateRange,
    PerpetualPeriod,
    Platform,
    Property,
    RealCalendar,
    Relationship,
    ResponsibleParty,
    Standard,
    StandardName,
    Component,
    ComponentLanguage,
    ComponentLanguageProperty,
    ComponentProperty,
    Composition,
    Connection,
    ConnectionEndpoint,
    ConnectionProperty,
    Coupling,
    CouplingEndpoint,
    CouplingProperty,
    Deployment,
    EntryPoint,
    ModelComponent,
    Parallelisation,
    ProcessorComponent,
    Rank,
    SpatialRegridding,
    SpatialRegriddingProperty,
    SpatialRegriddingUserMethod,
    StatisticalModelComponent,
    TimeLag,
    TimeTransformation,
    Timing,
    # Enums
    ConformanceType,
    DownscalingType,
    EnsembleType,
    ExperimentRelationshipType,
    FrequencyType,
    ProjectType,
    ResolutionType,
    SimulationRelationshipType,
    SimulationType,
    DataHierarchyType,
    DataStatusType,
    ArcTypeEnum,
    DiscretizationEnum,
    FeatureTypeEnum,
    GeometryTypeEnum,
    GridNodePositionEnum,
    GridTypeEnum,
    HorizontalCsEnum,
    RefinementTypeEnum,
    VerticalCsEnum,
    CimFeatureType,
    CimResultType,
    CimScopeCodeType,
    QualityIssueType,
    QualitySeverityType,
    QualityStatusType,
    ChangePropertyType,
    CompilerType,
    DataPurpose,
    InterconnectType,
    MachineType,
    MachineVendorType,
    OperatingSystemType,
    ProcessorType,
    UnitType,
    ComponentPropertyIntentType,
    ConnectionType,
    CouplingFrameworkType,
    ModelComponentType,
    SpatialRegriddingDimensionType,
    SpatialRegriddingStandardMethodType,
    StatisticalModelComponentType,
    TimeMappingType,
    TimingUnits,
]




# -*- coding: utf-8 -*-

"""
.. module:: misc_document_set.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.misc.document_set document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _propogate_institute,
        _propogate_meta_info,
        _propogate_ext_info,
        )


def unpack_children(ctx):
    """Unpacks child documents.

    """
    children = [
        ctx.doc.experiment,
        ctx.doc.model,
        ctx.doc.platform,
        ctx.doc.simulation
        ]
    children += ctx.doc.data
    children += ctx.doc.ensembles
    children += ctx.doc.grids

    ctx.ext.children = [c for c in children if c]


def _propogate_meta_info(ctx):
    """Propogates document meta information.

    """
    if ctx.doc.simulation:
        simulation = ctx.doc.simulation
        ctx.meta.create_date = simulation.meta.create_date
        ctx.meta.external_ids = simulation.meta.external_ids
        ctx.meta.id = simulation.meta.id
        ctx.meta.version = simulation.meta.version


def _propogate_ext_info(ctx):
    """Propogates document extension information.

    """
    if ctx.doc.simulation:
        simulation = ctx.doc.simulation
        for field in (
            'display_name',
            'description',
            'summary_fields',
            'full_display_name'
            ):
            setattr(ctx.ext, field, getattr(simulation.ext, field))


def _propogate_institute(ctx):
    """Propogates institute code.

    """
    for child in ctx.ext.children:
        if not child.meta.institute:
            child.meta.institute = ctx.doc.meta.institute
